"use strict";

var _interopRequireWildcard = require("@babel/runtime/helpers/interopRequireWildcard");

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.OlmDevice = OlmDevice;
exports.WITHHELD_MESSAGES = void 0;

var _regenerator = _interopRequireDefault(require("@babel/runtime/regenerator"));

var _asyncToGenerator2 = _interopRequireDefault(require("@babel/runtime/helpers/asyncToGenerator"));

var _logger = require("../logger");

var _indexeddbCryptoStore = require("./store/indexeddb-crypto-store");

var algorithms = _interopRequireWildcard(require("./algorithms"));

function _createForOfIteratorHelper(o, allowArrayLike) { var it; if (typeof Symbol === "undefined" || o[Symbol.iterator] == null) { if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") { if (it) o = it; var i = 0; var F = function F() {}; return { s: F, n: function n() { if (i >= o.length) return { done: true }; return { done: false, value: o[i++] }; }, e: function (_e) { function e(_x61) { return _e.apply(this, arguments); } e.toString = function () { return _e.toString(); }; return e; }(function (e) { throw e; }), f: F }; } throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); } var normalCompletion = true, didErr = false, err; return { s: function s() { it = o[Symbol.iterator](); }, n: function n() { var step = it.next(); normalCompletion = step.done; return step; }, e: function (_e2) { function e(_x62) { return _e2.apply(this, arguments); } e.toString = function () { return _e2.toString(); }; return e; }(function (e) { didErr = true; err = e; }), f: function f() { try { if (!normalCompletion && it["return"] != null) it["return"](); } finally { if (didErr) throw err; } } }; }

function _unsupportedIterableToArray(o, minLen) { if (!o) return; if (typeof o === "string") return _arrayLikeToArray(o, minLen); var n = Object.prototype.toString.call(o).slice(8, -1); if (n === "Object" && o.constructor) n = o.constructor.name; if (n === "Map" || n === "Set") return Array.from(o); if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen); }

function _arrayLikeToArray(arr, len) { if (len == null || len > arr.length) len = arr.length; for (var i = 0, arr2 = new Array(len); i < len; i++) { arr2[i] = arr[i]; } return arr2; }

// The maximum size of an event is 65K, and we base64 the content, so this is a
// reasonable approximation to the biggest plaintext we can encrypt.
var MAX_PLAINTEXT_LENGTH = 65536 * 3 / 4;

function checkPayloadLength(payloadString) {
  if (payloadString === undefined) {
    throw new Error("payloadString undefined");
  }

  if (payloadString.length > MAX_PLAINTEXT_LENGTH) {
    // might as well fail early here rather than letting the olm library throw
    // a cryptic memory allocation error.
    //
    // Note that even if we manage to do the encryption, the message send may fail,
    // because by the time we've wrapped the ciphertext in the event object, it may
    // exceed 65K. But at least we won't just fail with "abort()" in that case.
    var err = new Error("Message too long (" + payloadString.length + " bytes). " + "The maximum for an encrypted message is " + MAX_PLAINTEXT_LENGTH + " bytes."); // TODO: [TypeScript] We should have our own error types

    err.data = {
      errcode: "M_TOO_LARGE",
      error: "Payload too large for encrypted message"
    };
    throw err;
  }
}
/**
 * The type of object we use for importing and exporting megolm session data.
 *
 * @typedef {Object} module:crypto/OlmDevice.MegolmSessionData
 * @property {String} sender_key  Sender's Curve25519 device key
 * @property {String[]} forwarding_curve25519_key_chain Devices which forwarded
 *     this session to us (normally empty).
 * @property {Object<string, string>} sender_claimed_keys Other keys the sender claims.
 * @property {String} room_id     Room this session is used in
 * @property {String} session_id  Unique id for the session
 * @property {String} session_key Base64'ed key data
 */

/**
 * Manages the olm cryptography functions. Each OlmDevice has a single
 * OlmAccount and a number of OlmSessions.
 *
 * Accounts and sessions are kept pickled in the cryptoStore.
 *
 * @constructor
 * @alias module:crypto/OlmDevice
 *
 * @param {Object} cryptoStore A store for crypto data
 *
 * @property {string} deviceCurve25519Key   Curve25519 key for the account
 * @property {string} deviceEd25519Key      Ed25519 key for the account
 */


function OlmDevice(cryptoStore) {
  this._cryptoStore = cryptoStore;
  this._pickleKey = "DEFAULT_KEY"; // don't know these until we load the account from storage in init()

  this.deviceCurve25519Key = null;
  this.deviceEd25519Key = null;
  this._maxOneTimeKeys = null; // we don't bother stashing outboundgroupsessions in the cryptoStore -
  // instead we keep them here.

  this._outboundGroupSessionStore = {}; // Store a set of decrypted message indexes for each group session.
  // This partially mitigates a replay attack where a MITM resends a group
  // message into the room.
  //
  // When we decrypt a message and the message index matches a previously
  // decrypted message, one possible cause of that is that we are decrypting
  // the same event, and may not indicate an actual replay attack.  For
  // example, this could happen if we receive events, forget about them, and
  // then re-fetch them when we backfill.  So we store the event ID and
  // timestamp corresponding to each message index when we first decrypt it,
  // and compare these against the event ID and timestamp every time we use
  // that same index.  If they match, then we're probably decrypting the same
  // event and we don't consider it a replay attack.
  //
  // Keys are strings of form "<senderKey>|<session_id>|<message_index>"
  // Values are objects of the form "{id: <event id>, timestamp: <ts>}"

  this._inboundGroupSessionMessageIndexes = {}; // Keep track of sessions that we're starting, so that we don't start
  // multiple sessions for the same device at the same time.

  this._sessionsInProgress = {}; // Used by olm to serialise prekey message decryptions

  this._olmPrekeyPromise = Promise.resolve();
}
/**
 * Initialise the OlmAccount. This must be called before any other operations
 * on the OlmDevice.
 *
 * Data from an exported Olm device can be provided
 * in order to re-create this device.
 *
 * Attempts to load the OlmAccount from the crypto store, or creates one if none is
 * found.
 *
 * Reads the device keys from the OlmAccount object.
 *
 * @param {object} opts
 * @param {object} opts.fromExportedDevice (Optional) data from exported device
 *     that must be re-created.
 *     If present, opts.pickleKey is ignored
 *     (exported data already provides a pickle key)
 * @param {object} opts.pickleKey (Optional) pickle key to set instead of default one
 */


OlmDevice.prototype.init = /*#__PURE__*/(0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee() {
  var opts,
      e2eKeys,
      account,
      pickleKey,
      fromExportedDevice,
      _args = arguments;
  return _regenerator["default"].wrap(function _callee$(_context) {
    while (1) {
      switch (_context.prev = _context.next) {
        case 0:
          opts = _args.length > 0 && _args[0] !== undefined ? _args[0] : {};
          account = new global.Olm.Account();
          pickleKey = opts.pickleKey, fromExportedDevice = opts.fromExportedDevice;
          _context.prev = 3;

          if (!fromExportedDevice) {
            _context.next = 11;
            break;
          }

          if (pickleKey) {
            _logger.logger.warn('ignoring opts.pickleKey' + ' because opts.fromExportedDevice is present.');
          }

          this._pickleKey = fromExportedDevice.pickleKey;
          _context.next = 9;
          return _initialiseFromExportedDevice(fromExportedDevice, this._cryptoStore, this._pickleKey, account);

        case 9:
          _context.next = 14;
          break;

        case 11:
          if (pickleKey) {
            this._pickleKey = pickleKey;
          }

          _context.next = 14;
          return _initialiseAccount(this._cryptoStore, this._pickleKey, account);

        case 14:
          e2eKeys = JSON.parse(account.identity_keys());
          this._maxOneTimeKeys = account.max_number_of_one_time_keys();

        case 16:
          _context.prev = 16;
          account.free();
          return _context.finish(16);

        case 19:
          this.deviceCurve25519Key = e2eKeys.curve25519;
          this.deviceEd25519Key = e2eKeys.ed25519;

        case 21:
        case "end":
          return _context.stop();
      }
    }
  }, _callee, this, [[3,, 16, 19]]);
}));
/**
 * Populates the crypto store using data that was exported from an existing device.
 * Note that for now only the “account” and “sessions” stores are populated;
 * Other stores will be as with a new device.
 *
 * @param {Object} exportedData Data exported from another device
 *     through the “export” method.
 * @param {module:crypto/store/base~CryptoStore} cryptoStore storage for the crypto layer
 * @param {string} pickleKey the key that was used to pickle the exported data
 * @param {Olm.Account} account an olm account to initialize
 */

function _initialiseFromExportedDevice(_x, _x2, _x3, _x4) {
  return _initialiseFromExportedDevice2.apply(this, arguments);
}

function _initialiseFromExportedDevice2() {
  _initialiseFromExportedDevice2 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee24(exportedData, cryptoStore, pickleKey, account) {
    return _regenerator["default"].wrap(function _callee24$(_context24) {
      while (1) {
        switch (_context24.prev = _context24.next) {
          case 0:
            _context24.next = 2;
            return cryptoStore.doTxn('readwrite', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_ACCOUNT, _indexeddbCryptoStore.IndexedDBCryptoStore.STORE_SESSIONS], function (txn) {
              cryptoStore.storeAccount(txn, exportedData.pickledAccount);
              exportedData.sessions.forEach(function (session) {
                var deviceKey = session.deviceKey,
                    sessionId = session.sessionId;
                var sessionInfo = {
                  session: session.session,
                  lastReceivedMessageTs: session.lastReceivedMessageTs
                };
                cryptoStore.storeEndToEndSession(deviceKey, sessionId, sessionInfo, txn);
              });
            });

          case 2:
            account.unpickle(pickleKey, exportedData.pickledAccount);

          case 3:
          case "end":
            return _context24.stop();
        }
      }
    }, _callee24);
  }));
  return _initialiseFromExportedDevice2.apply(this, arguments);
}

function _initialiseAccount(_x5, _x6, _x7) {
  return _initialiseAccount2.apply(this, arguments);
}
/**
 * @return {array} The version of Olm.
 */


function _initialiseAccount2() {
  _initialiseAccount2 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee25(cryptoStore, pickleKey, account) {
    return _regenerator["default"].wrap(function _callee25$(_context25) {
      while (1) {
        switch (_context25.prev = _context25.next) {
          case 0:
            _context25.next = 2;
            return cryptoStore.doTxn('readwrite', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_ACCOUNT], function (txn) {
              cryptoStore.getAccount(txn, function (pickledAccount) {
                if (pickledAccount !== null) {
                  account.unpickle(pickleKey, pickledAccount);
                } else {
                  account.create();
                  pickledAccount = account.pickle(pickleKey);
                  cryptoStore.storeAccount(txn, pickledAccount);
                }
              });
            });

          case 2:
          case "end":
            return _context25.stop();
        }
      }
    }, _callee25);
  }));
  return _initialiseAccount2.apply(this, arguments);
}

OlmDevice.getOlmVersion = function () {
  return global.Olm.get_library_version();
};
/**
 * extract our OlmAccount from the crypto store and call the given function
 * with the account object
 * The `account` object is useable only within the callback passed to this
 * function and will be freed as soon the callback returns. It is *not*
 * useable for the rest of the lifetime of the transaction.
 * This function requires a live transaction object from cryptoStore.doTxn()
 * and therefore may only be called in a doTxn() callback.
 *
 * @param {*} txn Opaque transaction object from cryptoStore.doTxn()
 * @param {function} func
 * @private
 */


OlmDevice.prototype._getAccount = function (txn, func) {
  var _this = this;

  this._cryptoStore.getAccount(txn, function (pickledAccount) {
    var account = new global.Olm.Account();

    try {
      account.unpickle(_this._pickleKey, pickledAccount);
      func(account);
    } finally {
      account.free();
    }
  });
};
/*
 * Saves an account to the crypto store.
 * This function requires a live transaction object from cryptoStore.doTxn()
 * and therefore may only be called in a doTxn() callback.
 *
 * @param {*} txn Opaque transaction object from cryptoStore.doTxn()
 * @param {object} Olm.Account object
 * @private
 */


OlmDevice.prototype._storeAccount = function (txn, account) {
  this._cryptoStore.storeAccount(txn, account.pickle(this._pickleKey));
};
/**
 * Export data for re-creating the Olm device later.
 * TODO export data other than just account and (P2P) sessions.
 *
 * @return {Promise<object>} The exported data
*/


OlmDevice.prototype["export"] = /*#__PURE__*/(0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee2() {
  var _this2 = this;

  var result;
  return _regenerator["default"].wrap(function _callee2$(_context2) {
    while (1) {
      switch (_context2.prev = _context2.next) {
        case 0:
          result = {
            pickleKey: this._pickleKey
          };
          _context2.next = 3;
          return this._cryptoStore.doTxn('readonly', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_ACCOUNT, _indexeddbCryptoStore.IndexedDBCryptoStore.STORE_SESSIONS], function (txn) {
            _this2._cryptoStore.getAccount(txn, function (pickledAccount) {
              result.pickledAccount = pickledAccount;
            });

            result.sessions = []; // Note that the pickledSession object we get in the callback
            // is not exactly the same thing you get in method _getSession
            // see documentation of IndexedDBCryptoStore.getAllEndToEndSessions

            _this2._cryptoStore.getAllEndToEndSessions(txn, function (pickledSession) {
              result.sessions.push(pickledSession);
            });
          });

        case 3:
          return _context2.abrupt("return", result);

        case 4:
        case "end":
          return _context2.stop();
      }
    }
  }, _callee2, this);
}));
/**
 * extract an OlmSession from the session store and call the given function
 * The session is useable only within the callback passed to this
 * function and will be freed as soon the callback returns. It is *not*
 * useable for the rest of the lifetime of the transaction.
 *
 * @param {string} deviceKey
 * @param {string} sessionId
 * @param {*} txn Opaque transaction object from cryptoStore.doTxn()
 * @param {function} func
 * @private
 */

OlmDevice.prototype._getSession = function (deviceKey, sessionId, txn, func) {
  var _this3 = this;

  this._cryptoStore.getEndToEndSession(deviceKey, sessionId, txn, function (sessionInfo) {
    _this3._unpickleSession(sessionInfo, func);
  });
};
/**
 * Creates a session object from a session pickle and executes the given
 * function with it. The session object is destroyed once the function
 * returns.
 *
 * @param {object} sessionInfo
 * @param {function} func
 * @private
 */


OlmDevice.prototype._unpickleSession = function (sessionInfo, func) {
  var session = new global.Olm.Session();

  try {
    session.unpickle(this._pickleKey, sessionInfo.session);
    var unpickledSessInfo = Object.assign({}, sessionInfo, {
      session: session
    });
    func(unpickledSessInfo);
  } finally {
    session.free();
  }
};
/**
 * store our OlmSession in the session store
 *
 * @param {string} deviceKey
 * @param {object} sessionInfo {session: OlmSession, lastReceivedMessageTs: int}
 * @param {*} txn Opaque transaction object from cryptoStore.doTxn()
 * @private
 */


OlmDevice.prototype._saveSession = function (deviceKey, sessionInfo, txn) {
  var sessionId = sessionInfo.session.session_id();
  var pickledSessionInfo = Object.assign(sessionInfo, {
    session: sessionInfo.session.pickle(this._pickleKey)
  });

  this._cryptoStore.storeEndToEndSession(deviceKey, sessionId, pickledSessionInfo, txn);
};
/**
 * get an OlmUtility and call the given function
 *
 * @param {function} func
 * @return {object} result of func
 * @private
 */


OlmDevice.prototype._getUtility = function (func) {
  var utility = new global.Olm.Utility();

  try {
    return func(utility);
  } finally {
    utility.free();
  }
};
/**
 * Signs a message with the ed25519 key for this account.
 *
 * @param {string} message  message to be signed
 * @return {Promise<string>} base64-encoded signature
 */


OlmDevice.prototype.sign = /*#__PURE__*/function () {
  var _ref3 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee3(message) {
    var _this4 = this;

    var result;
    return _regenerator["default"].wrap(function _callee3$(_context3) {
      while (1) {
        switch (_context3.prev = _context3.next) {
          case 0:
            _context3.next = 2;
            return this._cryptoStore.doTxn('readonly', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_ACCOUNT], function (txn) {
              _this4._getAccount(txn, function (account) {
                result = account.sign(message);
              });
            });

          case 2:
            return _context3.abrupt("return", result);

          case 3:
          case "end":
            return _context3.stop();
        }
      }
    }, _callee3, this);
  }));

  return function (_x8) {
    return _ref3.apply(this, arguments);
  };
}();
/**
 * Get the current (unused, unpublished) one-time keys for this account.
 *
 * @return {object} one time keys; an object with the single property
 * <tt>curve25519</tt>, which is itself an object mapping key id to Curve25519
 * key.
 */


OlmDevice.prototype.getOneTimeKeys = /*#__PURE__*/(0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee4() {
  var _this5 = this;

  var result;
  return _regenerator["default"].wrap(function _callee4$(_context4) {
    while (1) {
      switch (_context4.prev = _context4.next) {
        case 0:
          _context4.next = 2;
          return this._cryptoStore.doTxn('readonly', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_ACCOUNT], function (txn) {
            _this5._getAccount(txn, function (account) {
              result = JSON.parse(account.one_time_keys());
            });
          });

        case 2:
          return _context4.abrupt("return", result);

        case 3:
        case "end":
          return _context4.stop();
      }
    }
  }, _callee4, this);
}));
/**
 * Get the maximum number of one-time keys we can store.
 *
 * @return {number} number of keys
 */

OlmDevice.prototype.maxNumberOfOneTimeKeys = function () {
  return this._maxOneTimeKeys;
};
/**
 * Marks all of the one-time keys as published.
 */


OlmDevice.prototype.markKeysAsPublished = /*#__PURE__*/(0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee5() {
  var _this6 = this;

  return _regenerator["default"].wrap(function _callee5$(_context5) {
    while (1) {
      switch (_context5.prev = _context5.next) {
        case 0:
          _context5.next = 2;
          return this._cryptoStore.doTxn('readwrite', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_ACCOUNT], function (txn) {
            _this6._getAccount(txn, function (account) {
              account.mark_keys_as_published();

              _this6._storeAccount(txn, account);
            });
          });

        case 2:
        case "end":
          return _context5.stop();
      }
    }
  }, _callee5, this);
}));
/**
 * Generate some new one-time keys
 *
 * @param {number} numKeys number of keys to generate
 * @return {Promise} Resolved once the account is saved back having generated the keys
 */

OlmDevice.prototype.generateOneTimeKeys = function (numKeys) {
  var _this7 = this;

  return this._cryptoStore.doTxn('readwrite', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_ACCOUNT], function (txn) {
    _this7._getAccount(txn, function (account) {
      account.generate_one_time_keys(numKeys);

      _this7._storeAccount(txn, account);
    });
  });
};
/**
 * Generate a new fallback keys
 *
 * @return {Promise} Resolved once the account is saved back having generated the key
 */


OlmDevice.prototype.generateFallbackKey = /*#__PURE__*/(0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee6() {
  var _this8 = this;

  return _regenerator["default"].wrap(function _callee6$(_context6) {
    while (1) {
      switch (_context6.prev = _context6.next) {
        case 0:
          _context6.next = 2;
          return this._cryptoStore.doTxn('readwrite', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_ACCOUNT], function (txn) {
            _this8._getAccount(txn, function (account) {
              account.generate_fallback_key();

              _this8._storeAccount(txn, account);
            });
          });

        case 2:
        case "end":
          return _context6.stop();
      }
    }
  }, _callee6, this);
}));
OlmDevice.prototype.getFallbackKey = /*#__PURE__*/(0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee7() {
  var _this9 = this;

  var result;
  return _regenerator["default"].wrap(function _callee7$(_context7) {
    while (1) {
      switch (_context7.prev = _context7.next) {
        case 0:
          _context7.next = 2;
          return this._cryptoStore.doTxn('readonly', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_ACCOUNT], function (txn) {
            _this9._getAccount(txn, function (account) {
              result = JSON.parse(account.fallback_key());
            });
          });

        case 2:
          return _context7.abrupt("return", result);

        case 3:
        case "end":
          return _context7.stop();
      }
    }
  }, _callee7, this);
}));
/**
 * Generate a new outbound session
 *
 * The new session will be stored in the cryptoStore.
 *
 * @param {string} theirIdentityKey remote user's Curve25519 identity key
 * @param {string} theirOneTimeKey  remote user's one-time Curve25519 key
 * @return {string} sessionId for the outbound session.
 */

OlmDevice.prototype.createOutboundSession = /*#__PURE__*/function () {
  var _ref8 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee8(theirIdentityKey, theirOneTimeKey) {
    var _this10 = this;

    var newSessionId;
    return _regenerator["default"].wrap(function _callee8$(_context8) {
      while (1) {
        switch (_context8.prev = _context8.next) {
          case 0:
            _context8.next = 2;
            return this._cryptoStore.doTxn('readwrite', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_ACCOUNT, _indexeddbCryptoStore.IndexedDBCryptoStore.STORE_SESSIONS], function (txn) {
              _this10._getAccount(txn, function (account) {
                var session = new global.Olm.Session();

                try {
                  session.create_outbound(account, theirIdentityKey, theirOneTimeKey);
                  newSessionId = session.session_id();

                  _this10._storeAccount(txn, account);

                  var sessionInfo = {
                    session: session,
                    // Pretend we've received a message at this point, otherwise
                    // if we try to send a message to the device, it won't use
                    // this session
                    lastReceivedMessageTs: Date.now()
                  };

                  _this10._saveSession(theirIdentityKey, sessionInfo, txn);
                } finally {
                  session.free();
                }
              });
            });

          case 2:
            return _context8.abrupt("return", newSessionId);

          case 3:
          case "end":
            return _context8.stop();
        }
      }
    }, _callee8, this);
  }));

  return function (_x9, _x10) {
    return _ref8.apply(this, arguments);
  };
}();
/**
 * Generate a new inbound session, given an incoming message
 *
 * @param {string} theirDeviceIdentityKey remote user's Curve25519 identity key
 * @param {number} messageType  messageType field from the received message (must be 0)
 * @param {string} ciphertext base64-encoded body from the received message
 *
 * @return {{payload: string, session_id: string}} decrypted payload, and
 *     session id of new session
 *
 * @raises {Error} if the received message was not valid (for instance, it
 *     didn't use a valid one-time key).
 */


OlmDevice.prototype.createInboundSession = /*#__PURE__*/function () {
  var _ref9 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee9(theirDeviceIdentityKey, messageType, ciphertext) {
    var _this11 = this;

    var result;
    return _regenerator["default"].wrap(function _callee9$(_context9) {
      while (1) {
        switch (_context9.prev = _context9.next) {
          case 0:
            if (!(messageType !== 0)) {
              _context9.next = 2;
              break;
            }

            throw new Error("Need messageType == 0 to create inbound session");

          case 2:
            _context9.next = 4;
            return this._cryptoStore.doTxn('readwrite', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_ACCOUNT, _indexeddbCryptoStore.IndexedDBCryptoStore.STORE_SESSIONS], function (txn) {
              _this11._getAccount(txn, function (account) {
                var session = new global.Olm.Session();

                try {
                  session.create_inbound_from(account, theirDeviceIdentityKey, ciphertext);
                  account.remove_one_time_keys(session);

                  _this11._storeAccount(txn, account);

                  var payloadString = session.decrypt(messageType, ciphertext);
                  var sessionInfo = {
                    session: session,
                    // this counts as a received message: set last received message time
                    // to now
                    lastReceivedMessageTs: Date.now()
                  };

                  _this11._saveSession(theirDeviceIdentityKey, sessionInfo, txn);

                  result = {
                    payload: payloadString,
                    session_id: session.session_id()
                  };
                } finally {
                  session.free();
                }
              });
            });

          case 4:
            return _context9.abrupt("return", result);

          case 5:
          case "end":
            return _context9.stop();
        }
      }
    }, _callee9, this);
  }));

  return function (_x11, _x12, _x13) {
    return _ref9.apply(this, arguments);
  };
}();
/**
 * Get a list of known session IDs for the given device
 *
 * @param {string} theirDeviceIdentityKey Curve25519 identity key for the
 *     remote device
 * @return {Promise<string[]>}  a list of known session ids for the device
 */


OlmDevice.prototype.getSessionIdsForDevice = /*#__PURE__*/function () {
  var _ref10 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee10(theirDeviceIdentityKey) {
    var _this12 = this;

    var sessionIds;
    return _regenerator["default"].wrap(function _callee10$(_context10) {
      while (1) {
        switch (_context10.prev = _context10.next) {
          case 0:
            if (!this._sessionsInProgress[theirDeviceIdentityKey]) {
              _context10.next = 9;
              break;
            }

            _logger.logger.log("waiting for olm session to be created");

            _context10.prev = 2;
            _context10.next = 5;
            return this._sessionsInProgress[theirDeviceIdentityKey];

          case 5:
            _context10.next = 9;
            break;

          case 7:
            _context10.prev = 7;
            _context10.t0 = _context10["catch"](2);

          case 9:
            _context10.next = 11;
            return this._cryptoStore.doTxn('readonly', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_SESSIONS], function (txn) {
              _this12._cryptoStore.getEndToEndSessions(theirDeviceIdentityKey, txn, function (sessions) {
                sessionIds = Object.keys(sessions);
              });
            });

          case 11:
            return _context10.abrupt("return", sessionIds);

          case 12:
          case "end":
            return _context10.stop();
        }
      }
    }, _callee10, this, [[2, 7]]);
  }));

  return function (_x14) {
    return _ref10.apply(this, arguments);
  };
}();
/**
 * Get the right olm session id for encrypting messages to the given identity key
 *
 * @param {string} theirDeviceIdentityKey Curve25519 identity key for the
 *     remote device
 * @param {boolean} nowait Don't wait for an in-progress session to complete.
 *     This should only be set to true of the calling function is the function
 *     that marked the session as being in-progress.
 * @return {Promise<?string>}  session id, or null if no established session
 */


OlmDevice.prototype.getSessionIdForDevice = /*#__PURE__*/function () {
  var _ref11 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee11(theirDeviceIdentityKey, nowait) {
    var sessionInfos, idxOfBest, i, thisSessInfo, thisLastReceived, bestSessInfo, bestLastReceived;
    return _regenerator["default"].wrap(function _callee11$(_context11) {
      while (1) {
        switch (_context11.prev = _context11.next) {
          case 0:
            _context11.next = 2;
            return this.getSessionInfoForDevice(theirDeviceIdentityKey, nowait);

          case 2:
            sessionInfos = _context11.sent;

            if (!(sessionInfos.length === 0)) {
              _context11.next = 5;
              break;
            }

            return _context11.abrupt("return", null);

          case 5:
            // Use the session that has most recently received a message
            idxOfBest = 0;

            for (i = 1; i < sessionInfos.length; i++) {
              thisSessInfo = sessionInfos[i];
              thisLastReceived = thisSessInfo.lastReceivedMessageTs === undefined ? 0 : thisSessInfo.lastReceivedMessageTs;
              bestSessInfo = sessionInfos[idxOfBest];
              bestLastReceived = bestSessInfo.lastReceivedMessageTs === undefined ? 0 : bestSessInfo.lastReceivedMessageTs;

              if (thisLastReceived > bestLastReceived || thisLastReceived === bestLastReceived && thisSessInfo.sessionId < bestSessInfo.sessionId) {
                idxOfBest = i;
              }
            }

            return _context11.abrupt("return", sessionInfos[idxOfBest].sessionId);

          case 8:
          case "end":
            return _context11.stop();
        }
      }
    }, _callee11, this);
  }));

  return function (_x15, _x16) {
    return _ref11.apply(this, arguments);
  };
}();
/**
 * Get information on the active Olm sessions for a device.
 * <p>
 * Returns an array, with an entry for each active session. The first entry in
 * the result will be the one used for outgoing messages. Each entry contains
 * the keys 'hasReceivedMessage' (true if the session has received an incoming
 * message and is therefore past the pre-key stage), and 'sessionId'.
 *
 * @param {string} deviceIdentityKey Curve25519 identity key for the device
 * @param {boolean} nowait Don't wait for an in-progress session to complete.
 *     This should only be set to true of the calling function is the function
 *     that marked the session as being in-progress.
 * @return {Array.<{sessionId: string, hasReceivedMessage: Boolean}>}
 */


OlmDevice.prototype.getSessionInfoForDevice = /*#__PURE__*/function () {
  var _ref12 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee12(deviceIdentityKey, nowait) {
    var _this13 = this;

    var info;
    return _regenerator["default"].wrap(function _callee12$(_context12) {
      while (1) {
        switch (_context12.prev = _context12.next) {
          case 0:
            if (!(this._sessionsInProgress[deviceIdentityKey] && !nowait)) {
              _context12.next = 9;
              break;
            }

            _logger.logger.log("waiting for olm session to be created");

            _context12.prev = 2;
            _context12.next = 5;
            return this._sessionsInProgress[deviceIdentityKey];

          case 5:
            _context12.next = 9;
            break;

          case 7:
            _context12.prev = 7;
            _context12.t0 = _context12["catch"](2);

          case 9:
            info = [];
            _context12.next = 12;
            return this._cryptoStore.doTxn('readonly', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_SESSIONS], function (txn) {
              _this13._cryptoStore.getEndToEndSessions(deviceIdentityKey, txn, function (sessions) {
                var sessionIds = Object.keys(sessions).sort();

                var _iterator = _createForOfIteratorHelper(sessionIds),
                    _step;

                try {
                  var _loop = function _loop() {
                    var sessionId = _step.value;

                    _this13._unpickleSession(sessions[sessionId], function (sessInfo) {
                      info.push({
                        lastReceivedMessageTs: sessInfo.lastReceivedMessageTs,
                        hasReceivedMessage: sessInfo.session.has_received_message(),
                        sessionId: sessionId
                      });
                    });
                  };

                  for (_iterator.s(); !(_step = _iterator.n()).done;) {
                    _loop();
                  }
                } catch (err) {
                  _iterator.e(err);
                } finally {
                  _iterator.f();
                }
              });
            });

          case 12:
            return _context12.abrupt("return", info);

          case 13:
          case "end":
            return _context12.stop();
        }
      }
    }, _callee12, this, [[2, 7]]);
  }));

  return function (_x17, _x18) {
    return _ref12.apply(this, arguments);
  };
}();
/**
 * Encrypt an outgoing message using an existing session
 *
 * @param {string} theirDeviceIdentityKey Curve25519 identity key for the
 *     remote device
 * @param {string} sessionId  the id of the active session
 * @param {string} payloadString  payload to be encrypted and sent
 *
 * @return {Promise<string>} ciphertext
 */


OlmDevice.prototype.encryptMessage = /*#__PURE__*/function () {
  var _ref13 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee13(theirDeviceIdentityKey, sessionId, payloadString) {
    var _this14 = this;

    var res;
    return _regenerator["default"].wrap(function _callee13$(_context13) {
      while (1) {
        switch (_context13.prev = _context13.next) {
          case 0:
            checkPayloadLength(payloadString);
            _context13.next = 3;
            return this._cryptoStore.doTxn('readwrite', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_SESSIONS], function (txn) {
              _this14._getSession(theirDeviceIdentityKey, sessionId, txn, function (sessionInfo) {
                var sessionDesc = sessionInfo.session.describe();

                _logger.logger.log("encryptMessage: Olm Session ID " + sessionId + " to " + theirDeviceIdentityKey + ": " + sessionDesc);

                res = sessionInfo.session.encrypt(payloadString);

                _this14._saveSession(theirDeviceIdentityKey, sessionInfo, txn);
              });
            });

          case 3:
            return _context13.abrupt("return", res);

          case 4:
          case "end":
            return _context13.stop();
        }
      }
    }, _callee13, this);
  }));

  return function (_x19, _x20, _x21) {
    return _ref13.apply(this, arguments);
  };
}();
/**
 * Decrypt an incoming message using an existing session
 *
 * @param {string} theirDeviceIdentityKey Curve25519 identity key for the
 *     remote device
 * @param {string} sessionId  the id of the active session
 * @param {number} messageType  messageType field from the received message
 * @param {string} ciphertext base64-encoded body from the received message
 *
 * @return {Promise<string>} decrypted payload.
 */


OlmDevice.prototype.decryptMessage = /*#__PURE__*/function () {
  var _ref14 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee14(theirDeviceIdentityKey, sessionId, messageType, ciphertext) {
    var _this15 = this;

    var payloadString;
    return _regenerator["default"].wrap(function _callee14$(_context14) {
      while (1) {
        switch (_context14.prev = _context14.next) {
          case 0:
            _context14.next = 2;
            return this._cryptoStore.doTxn('readwrite', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_SESSIONS], function (txn) {
              _this15._getSession(theirDeviceIdentityKey, sessionId, txn, function (sessionInfo) {
                var sessionDesc = sessionInfo.session.describe();

                _logger.logger.log("decryptMessage: Olm Session ID " + sessionId + " from " + theirDeviceIdentityKey + ": " + sessionDesc);

                payloadString = sessionInfo.session.decrypt(messageType, ciphertext);
                sessionInfo.lastReceivedMessageTs = Date.now();

                _this15._saveSession(theirDeviceIdentityKey, sessionInfo, txn);
              });
            });

          case 2:
            return _context14.abrupt("return", payloadString);

          case 3:
          case "end":
            return _context14.stop();
        }
      }
    }, _callee14, this);
  }));

  return function (_x22, _x23, _x24, _x25) {
    return _ref14.apply(this, arguments);
  };
}();
/**
 * Determine if an incoming messages is a prekey message matching an existing session
 *
 * @param {string} theirDeviceIdentityKey Curve25519 identity key for the
 *     remote device
 * @param {string} sessionId  the id of the active session
 * @param {number} messageType  messageType field from the received message
 * @param {string} ciphertext base64-encoded body from the received message
 *
 * @return {Promise<boolean>} true if the received message is a prekey message which matches
 *    the given session.
 */


OlmDevice.prototype.matchesSession = /*#__PURE__*/function () {
  var _ref15 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee15(theirDeviceIdentityKey, sessionId, messageType, ciphertext) {
    var _this16 = this;

    var matches;
    return _regenerator["default"].wrap(function _callee15$(_context15) {
      while (1) {
        switch (_context15.prev = _context15.next) {
          case 0:
            if (!(messageType !== 0)) {
              _context15.next = 2;
              break;
            }

            return _context15.abrupt("return", false);

          case 2:
            _context15.next = 4;
            return this._cryptoStore.doTxn('readonly', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_SESSIONS], function (txn) {
              _this16._getSession(theirDeviceIdentityKey, sessionId, txn, function (sessionInfo) {
                matches = sessionInfo.session.matches_inbound(ciphertext);
              });
            });

          case 4:
            return _context15.abrupt("return", matches);

          case 5:
          case "end":
            return _context15.stop();
        }
      }
    }, _callee15, this);
  }));

  return function (_x26, _x27, _x28, _x29) {
    return _ref15.apply(this, arguments);
  };
}();

OlmDevice.prototype.recordSessionProblem = /*#__PURE__*/function () {
  var _ref16 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee16(deviceKey, type, fixed) {
    return _regenerator["default"].wrap(function _callee16$(_context16) {
      while (1) {
        switch (_context16.prev = _context16.next) {
          case 0:
            _context16.next = 2;
            return this._cryptoStore.storeEndToEndSessionProblem(deviceKey, type, fixed);

          case 2:
          case "end":
            return _context16.stop();
        }
      }
    }, _callee16, this);
  }));

  return function (_x30, _x31, _x32) {
    return _ref16.apply(this, arguments);
  };
}();

OlmDevice.prototype.sessionMayHaveProblems = /*#__PURE__*/function () {
  var _ref17 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee17(deviceKey, timestamp) {
    return _regenerator["default"].wrap(function _callee17$(_context17) {
      while (1) {
        switch (_context17.prev = _context17.next) {
          case 0:
            _context17.next = 2;
            return this._cryptoStore.getEndToEndSessionProblem(deviceKey, timestamp);

          case 2:
            return _context17.abrupt("return", _context17.sent);

          case 3:
          case "end":
            return _context17.stop();
        }
      }
    }, _callee17, this);
  }));

  return function (_x33, _x34) {
    return _ref17.apply(this, arguments);
  };
}();

OlmDevice.prototype.filterOutNotifiedErrorDevices = /*#__PURE__*/function () {
  var _ref18 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee18(devices) {
    return _regenerator["default"].wrap(function _callee18$(_context18) {
      while (1) {
        switch (_context18.prev = _context18.next) {
          case 0:
            _context18.next = 2;
            return this._cryptoStore.filterOutNotifiedErrorDevices(devices);

          case 2:
            return _context18.abrupt("return", _context18.sent);

          case 3:
          case "end":
            return _context18.stop();
        }
      }
    }, _callee18, this);
  }));

  return function (_x35) {
    return _ref18.apply(this, arguments);
  };
}(); // Outbound group session
// ======================

/**
 * store an OutboundGroupSession in _outboundGroupSessionStore
 *
 * @param {Olm.OutboundGroupSession} session
 * @private
 */


OlmDevice.prototype._saveOutboundGroupSession = function (session) {
  var pickledSession = session.pickle(this._pickleKey);
  this._outboundGroupSessionStore[session.session_id()] = pickledSession;
};
/**
 * extract an OutboundGroupSession from _outboundGroupSessionStore and call the
 * given function
 *
 * @param {string} sessionId
 * @param {function} func
 * @return {object} result of func
 * @private
 */


OlmDevice.prototype._getOutboundGroupSession = function (sessionId, func) {
  var pickled = this._outboundGroupSessionStore[sessionId];

  if (pickled === undefined) {
    throw new Error("Unknown outbound group session " + sessionId);
  }

  var session = new global.Olm.OutboundGroupSession();

  try {
    session.unpickle(this._pickleKey, pickled);
    return func(session);
  } finally {
    session.free();
  }
};
/**
 * Generate a new outbound group session
 *
 * @return {string} sessionId for the outbound session.
 */


OlmDevice.prototype.createOutboundGroupSession = function () {
  var session = new global.Olm.OutboundGroupSession();

  try {
    session.create();

    this._saveOutboundGroupSession(session);

    return session.session_id();
  } finally {
    session.free();
  }
};
/**
 * Encrypt an outgoing message with an outbound group session
 *
 * @param {string} sessionId  the id of the outboundgroupsession
 * @param {string} payloadString  payload to be encrypted and sent
 *
 * @return {string} ciphertext
 */


OlmDevice.prototype.encryptGroupMessage = function (sessionId, payloadString) {
  var self = this;

  _logger.logger.log("encrypting msg with megolm session ".concat(sessionId));

  checkPayloadLength(payloadString);
  return this._getOutboundGroupSession(sessionId, function (session) {
    var res = session.encrypt(payloadString);

    self._saveOutboundGroupSession(session);

    return res;
  });
};
/**
 * Get the session keys for an outbound group session
 *
 * @param {string} sessionId  the id of the outbound group session
 *
 * @return {{chain_index: number, key: string}} current chain index, and
 *     base64-encoded secret key.
 */


OlmDevice.prototype.getOutboundGroupSessionKey = function (sessionId) {
  return this._getOutboundGroupSession(sessionId, function (session) {
    return {
      chain_index: session.message_index(),
      key: session.session_key()
    };
  });
}; // Inbound group session
// =====================

/**
 * data stored in the session store about an inbound group session
 *
 * @typedef {Object} InboundGroupSessionData
 * @property {string} room_Id
 * @property {string} session   pickled Olm.InboundGroupSession
 * @property {Object<string, string>} keysClaimed
 * @property {Array<string>} forwardingCurve25519KeyChain  Devices involved in forwarding
 *     this session to us (normally empty).
 */

/**
 * Unpickle a session from a sessionData object and invoke the given function.
 * The session is valid only until func returns.
 *
 * @param {Object} sessionData Object describing the session.
 * @param {function(Olm.InboundGroupSession)} func Invoked with the unpickled session
 * @return {*} result of func
 */


OlmDevice.prototype._unpickleInboundGroupSession = function (sessionData, func) {
  var session = new global.Olm.InboundGroupSession();

  try {
    session.unpickle(this._pickleKey, sessionData.session);
    return func(session);
  } finally {
    session.free();
  }
};
/**
 * extract an InboundGroupSession from the crypto store and call the given function
 *
 * @param {string} roomId The room ID to extract the session for, or null to fetch
 *     sessions for any room.
 * @param {string} senderKey
 * @param {string} sessionId
 * @param {*} txn Opaque transaction object from cryptoStore.doTxn()
 * @param {function(Olm.InboundGroupSession, InboundGroupSessionData)} func
 *   function to call.
 *
 * @private
 */


OlmDevice.prototype._getInboundGroupSession = function (roomId, senderKey, sessionId, txn, func) {
  var _this17 = this;

  this._cryptoStore.getEndToEndInboundGroupSession(senderKey, sessionId, txn, function (sessionData, withheld) {
    if (sessionData === null) {
      func(null, null, withheld);
      return;
    } // if we were given a room ID, check that the it matches the original one for the session. This stops
    // the HS pretending a message was targeting a different room.


    if (roomId !== null && roomId !== sessionData.room_id) {
      throw new Error("Mismatched room_id for inbound group session (expected " + sessionData.room_id + ", was " + roomId + ")");
    }

    _this17._unpickleInboundGroupSession(sessionData, function (session) {
      func(session, sessionData, withheld);
    });
  });
};
/**
 * Add an inbound group session to the session store
 *
 * @param {string} roomId     room in which this session will be used
 * @param {string} senderKey  base64-encoded curve25519 key of the sender
 * @param {Array<string>} forwardingCurve25519KeyChain  Devices involved in forwarding
 *     this session to us.
 * @param {string} sessionId  session identifier
 * @param {string} sessionKey base64-encoded secret key
 * @param {Object<string, string>} keysClaimed Other keys the sender claims.
 * @param {boolean} exportFormat true if the megolm keys are in export format
 *    (ie, they lack an ed25519 signature)
 * @param {Object} [extraSessionData={}] any other data to be include with the session
 */


OlmDevice.prototype.addInboundGroupSession = /*#__PURE__*/function () {
  var _ref19 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee19(roomId, senderKey, forwardingCurve25519KeyChain, sessionId, sessionKey, keysClaimed, exportFormat) {
    var _this18 = this;

    var extraSessionData,
        _args19 = arguments;
    return _regenerator["default"].wrap(function _callee19$(_context19) {
      while (1) {
        switch (_context19.prev = _context19.next) {
          case 0:
            extraSessionData = _args19.length > 7 && _args19[7] !== undefined ? _args19[7] : {};
            _context19.next = 3;
            return this._cryptoStore.doTxn('readwrite', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_INBOUND_GROUP_SESSIONS, _indexeddbCryptoStore.IndexedDBCryptoStore.STORE_INBOUND_GROUP_SESSIONS_WITHHELD], function (txn) {
              /* if we already have this session, consider updating it */
              _this18._getInboundGroupSession(roomId, senderKey, sessionId, txn, function (existingSession, existingSessionData) {
                // new session.
                var session = new global.Olm.InboundGroupSession();

                try {
                  if (exportFormat) {
                    session.import_session(sessionKey);
                  } else {
                    session.create(sessionKey);
                  }

                  if (sessionId != session.session_id()) {
                    throw new Error("Mismatched group session ID from senderKey: " + senderKey);
                  }

                  if (existingSession) {
                    _logger.logger.log("Update for megolm session " + senderKey + "/" + sessionId);

                    if (existingSession.first_known_index() <= session.first_known_index()) {
                      // existing session has lower index (i.e. can
                      // decrypt more), so keep it
                      _logger.logger.log("Keeping existing megolm session ".concat(sessionId));

                      return;
                    }
                  }

                  _logger.logger.info("Storing megolm session " + senderKey + "/" + sessionId + " with first index " + session.first_known_index());

                  var sessionData = Object.assign({}, extraSessionData, {
                    room_id: roomId,
                    session: session.pickle(_this18._pickleKey),
                    keysClaimed: keysClaimed,
                    forwardingCurve25519KeyChain: forwardingCurve25519KeyChain
                  });

                  _this18._cryptoStore.storeEndToEndInboundGroupSession(senderKey, sessionId, sessionData, txn);
                } finally {
                  session.free();
                }
              });
            });

          case 3:
          case "end":
            return _context19.stop();
        }
      }
    }, _callee19, this);
  }));

  return function (_x36, _x37, _x38, _x39, _x40, _x41, _x42) {
    return _ref19.apply(this, arguments);
  };
}();
/**
 * Record in the data store why an inbound group session was withheld.
 *
 * @param {string} roomId     room that the session belongs to
 * @param {string} senderKey  base64-encoded curve25519 key of the sender
 * @param {string} sessionId  session identifier
 * @param {string} code       reason code
 * @param {string} reason     human-readable version of `code`
 */


OlmDevice.prototype.addInboundGroupSessionWithheld = /*#__PURE__*/function () {
  var _ref20 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee20(roomId, senderKey, sessionId, code, reason) {
    var _this19 = this;

    return _regenerator["default"].wrap(function _callee20$(_context20) {
      while (1) {
        switch (_context20.prev = _context20.next) {
          case 0:
            _context20.next = 2;
            return this._cryptoStore.doTxn('readwrite', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_INBOUND_GROUP_SESSIONS_WITHHELD], function (txn) {
              _this19._cryptoStore.storeEndToEndInboundGroupSessionWithheld(senderKey, sessionId, {
                room_id: roomId,
                code: code,
                reason: reason
              }, txn);
            });

          case 2:
          case "end":
            return _context20.stop();
        }
      }
    }, _callee20, this);
  }));

  return function (_x43, _x44, _x45, _x46, _x47) {
    return _ref20.apply(this, arguments);
  };
}();

var WITHHELD_MESSAGES = {
  "m.unverified": "The sender has disabled encrypting to unverified devices.",
  "m.blacklisted": "The sender has blocked you.",
  "m.unauthorised": "You are not authorised to read the message.",
  "m.no_olm": "Unable to establish a secure channel."
};
/**
 * Calculate the message to use for the exception when a session key is withheld.
 *
 * @param {object} withheld  An object that describes why the key was withheld.
 *
 * @return {string} the message
 *
 * @private
 */

exports.WITHHELD_MESSAGES = WITHHELD_MESSAGES;

function _calculateWithheldMessage(withheld) {
  if (withheld.code && withheld.code in WITHHELD_MESSAGES) {
    return WITHHELD_MESSAGES[withheld.code];
  } else if (withheld.reason) {
    return withheld.reason;
  } else {
    return "decryption key withheld";
  }
}
/**
 * Decrypt a received message with an inbound group session
 *
 * @param {string} roomId    room in which the message was received
 * @param {string} senderKey base64-encoded curve25519 key of the sender
 * @param {string} sessionId session identifier
 * @param {string} body      base64-encoded body of the encrypted message
 * @param {string} eventId   ID of the event being decrypted
 * @param {Number} timestamp timestamp of the event being decrypted
 *
 * @return {null} the sessionId is unknown
 *
 * @return {Promise<{result: string, senderKey: string,
 *    forwardingCurve25519KeyChain: Array<string>,
 *    keysClaimed: Object<string, string>}>}
 */


OlmDevice.prototype.decryptGroupMessage = /*#__PURE__*/function () {
  var _ref21 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee21(roomId, senderKey, sessionId, body, eventId, timestamp) {
    var _this20 = this;

    var result, error;
    return _regenerator["default"].wrap(function _callee21$(_context21) {
      while (1) {
        switch (_context21.prev = _context21.next) {
          case 0:
            _context21.next = 2;
            return this._cryptoStore.doTxn('readwrite', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_INBOUND_GROUP_SESSIONS, _indexeddbCryptoStore.IndexedDBCryptoStore.STORE_INBOUND_GROUP_SESSIONS_WITHHELD], function (txn) {
              _this20._getInboundGroupSession(roomId, senderKey, sessionId, txn, function (session, sessionData, withheld) {
                if (session === null) {
                  if (withheld) {
                    error = new algorithms.DecryptionError("MEGOLM_UNKNOWN_INBOUND_SESSION_ID", _calculateWithheldMessage(withheld), {
                      session: senderKey + '|' + sessionId
                    });
                  }

                  result = null;
                  return;
                }

                var res;

                try {
                  res = session.decrypt(body);
                } catch (e) {
                  if (e && e.message === 'OLM.UNKNOWN_MESSAGE_INDEX' && withheld) {
                    error = new algorithms.DecryptionError("MEGOLM_UNKNOWN_INBOUND_SESSION_ID", _calculateWithheldMessage(withheld), {
                      session: senderKey + '|' + sessionId
                    });
                  } else {
                    error = e;
                  }

                  return;
                }

                var plaintext = res.plaintext;

                if (plaintext === undefined) {
                  // Compatibility for older olm versions.
                  plaintext = res;
                } else {
                  // Check if we have seen this message index before to detect replay attacks.
                  // If the event ID and timestamp are specified, and the match the event ID
                  // and timestamp from the last time we used this message index, then we
                  // don't consider it a replay attack.
                  var messageIndexKey = senderKey + "|" + sessionId + "|" + res.message_index;

                  if (messageIndexKey in _this20._inboundGroupSessionMessageIndexes) {
                    var msgInfo = _this20._inboundGroupSessionMessageIndexes[messageIndexKey];

                    if (msgInfo.id !== eventId || msgInfo.timestamp !== timestamp) {
                      error = new Error("Duplicate message index, possible replay attack: " + messageIndexKey);
                      return;
                    }
                  }

                  _this20._inboundGroupSessionMessageIndexes[messageIndexKey] = {
                    id: eventId,
                    timestamp: timestamp
                  };
                }

                sessionData.session = session.pickle(_this20._pickleKey);

                _this20._cryptoStore.storeEndToEndInboundGroupSession(senderKey, sessionId, sessionData, txn);

                result = {
                  result: plaintext,
                  keysClaimed: sessionData.keysClaimed || {},
                  senderKey: senderKey,
                  forwardingCurve25519KeyChain: sessionData.forwardingCurve25519KeyChain || [],
                  untrusted: sessionData.untrusted
                };
              });
            });

          case 2:
            if (!error) {
              _context21.next = 4;
              break;
            }

            throw error;

          case 4:
            return _context21.abrupt("return", result);

          case 5:
          case "end":
            return _context21.stop();
        }
      }
    }, _callee21, this);
  }));

  return function (_x48, _x49, _x50, _x51, _x52, _x53) {
    return _ref21.apply(this, arguments);
  };
}();
/**
 * Determine if we have the keys for a given megolm session
 *
 * @param {string} roomId    room in which the message was received
 * @param {string} senderKey base64-encoded curve25519 key of the sender
 * @param {string} sessionId session identifier
 *
 * @returns {Promise<boolean>} true if we have the keys to this session
 */


OlmDevice.prototype.hasInboundSessionKeys = /*#__PURE__*/function () {
  var _ref22 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee22(roomId, senderKey, sessionId) {
    var _this21 = this;

    var result;
    return _regenerator["default"].wrap(function _callee22$(_context22) {
      while (1) {
        switch (_context22.prev = _context22.next) {
          case 0:
            _context22.next = 2;
            return this._cryptoStore.doTxn('readonly', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_INBOUND_GROUP_SESSIONS, _indexeddbCryptoStore.IndexedDBCryptoStore.STORE_INBOUND_GROUP_SESSIONS_WITHHELD], function (txn) {
              _this21._cryptoStore.getEndToEndInboundGroupSession(senderKey, sessionId, txn, function (sessionData) {
                if (sessionData === null) {
                  result = false;
                  return;
                }

                if (roomId !== sessionData.room_id) {
                  _logger.logger.warn("requested keys for inbound group session ".concat(senderKey, "|") + "".concat(sessionId, ", with incorrect room_id ") + "(expected ".concat(sessionData.room_id, ", ") + "was ".concat(roomId, ")"));

                  result = false;
                } else {
                  result = true;
                }
              });
            });

          case 2:
            return _context22.abrupt("return", result);

          case 3:
          case "end":
            return _context22.stop();
        }
      }
    }, _callee22, this);
  }));

  return function (_x54, _x55, _x56) {
    return _ref22.apply(this, arguments);
  };
}();
/**
 * Extract the keys to a given megolm session, for sharing
 *
 * @param {string} roomId    room in which the message was received
 * @param {string} senderKey base64-encoded curve25519 key of the sender
 * @param {string} sessionId session identifier
 * @param {integer} chainIndex The chain index at which to export the session.
 *     If omitted, export at the first index we know about.
 *
 * @returns {Promise<{chain_index: number, key: string,
 *        forwarding_curve25519_key_chain: Array<string>,
 *        sender_claimed_ed25519_key: string
 *    }>}
 *    details of the session key. The key is a base64-encoded megolm key in
 *    export format.
 *
 * @throws Error If the given chain index could not be obtained from the known
 *     index (ie. the given chain index is before the first we have).
 */


OlmDevice.prototype.getInboundGroupSessionKey = /*#__PURE__*/function () {
  var _ref23 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee23(roomId, senderKey, sessionId, chainIndex) {
    var _this22 = this;

    var result;
    return _regenerator["default"].wrap(function _callee23$(_context23) {
      while (1) {
        switch (_context23.prev = _context23.next) {
          case 0:
            _context23.next = 2;
            return this._cryptoStore.doTxn('readonly', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_INBOUND_GROUP_SESSIONS, _indexeddbCryptoStore.IndexedDBCryptoStore.STORE_INBOUND_GROUP_SESSIONS_WITHHELD], function (txn) {
              _this22._getInboundGroupSession(roomId, senderKey, sessionId, txn, function (session, sessionData) {
                if (session === null) {
                  result = null;
                  return;
                }

                if (chainIndex === undefined) {
                  chainIndex = session.first_known_index();
                }

                var exportedSession = session.export_session(chainIndex);
                var claimedKeys = sessionData.keysClaimed || {};
                var senderEd25519Key = claimedKeys.ed25519 || null;
                result = {
                  "chain_index": chainIndex,
                  "key": exportedSession,
                  "forwarding_curve25519_key_chain": sessionData.forwardingCurve25519KeyChain || [],
                  "sender_claimed_ed25519_key": senderEd25519Key
                };
              });
            });

          case 2:
            return _context23.abrupt("return", result);

          case 3:
          case "end":
            return _context23.stop();
        }
      }
    }, _callee23, this);
  }));

  return function (_x57, _x58, _x59, _x60) {
    return _ref23.apply(this, arguments);
  };
}();
/**
 * Export an inbound group session
 *
 * @param {string} senderKey base64-encoded curve25519 key of the sender
 * @param {string} sessionId session identifier
 * @param {string} sessionData The session object from the store
 * @return {module:crypto/OlmDevice.MegolmSessionData} exported session data
 */


OlmDevice.prototype.exportInboundGroupSession = function (senderKey, sessionId, sessionData) {
  return this._unpickleInboundGroupSession(sessionData, function (session) {
    var messageIndex = session.first_known_index();
    return {
      "sender_key": senderKey,
      "sender_claimed_keys": sessionData.keysClaimed,
      "room_id": sessionData.room_id,
      "session_id": sessionId,
      "session_key": session.export_session(messageIndex),
      "forwarding_curve25519_key_chain": session.forwardingCurve25519KeyChain || [],
      "first_known_index": session.first_known_index()
    };
  });
}; // Utilities
// =========

/**
 * Verify an ed25519 signature.
 *
 * @param {string} key ed25519 key
 * @param {string} message message which was signed
 * @param {string} signature base64-encoded signature to be checked
 *
 * @raises {Error} if there is a problem with the verification. If the key was
 * too small then the message will be "OLM.INVALID_BASE64". If the signature
 * was invalid then the message will be "OLM.BAD_MESSAGE_MAC".
 */


OlmDevice.prototype.verifySignature = function (key, message, signature) {
  this._getUtility(function (util) {
    util.ed25519_verify(key, message, signature);
  });
};