"use strict";

var _interopRequireWildcard = require("@babel/runtime/helpers/interopRequireWildcard");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.WebStorageSessionStore = WebStorageSessionStore;

var utils = _interopRequireWildcard(require("../../utils"));

var _logger = require("../../logger");

function _createForOfIteratorHelper(o, allowArrayLike) { var it; if (typeof Symbol === "undefined" || o[Symbol.iterator] == null) { if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") { if (it) o = it; var i = 0; var F = function F() {}; return { s: F, n: function n() { if (i >= o.length) return { done: true }; return { done: false, value: o[i++] }; }, e: function e(_e) { throw _e; }, f: F }; } throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); } var normalCompletion = true, didErr = false, err; return { s: function s() { it = o[Symbol.iterator](); }, n: function n() { var step = it.next(); normalCompletion = step.done; return step; }, e: function e(_e2) { didErr = true; err = _e2; }, f: function f() { try { if (!normalCompletion && it["return"] != null) it["return"](); } finally { if (didErr) throw err; } } }; }

function _unsupportedIterableToArray(o, minLen) { if (!o) return; if (typeof o === "string") return _arrayLikeToArray(o, minLen); var n = Object.prototype.toString.call(o).slice(8, -1); if (n === "Object" && o.constructor) n = o.constructor.name; if (n === "Map" || n === "Set") return Array.from(o); if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen); }

function _arrayLikeToArray(arr, len) { if (len == null || len > arr.length) len = arr.length; for (var i = 0, arr2 = new Array(len); i < len; i++) { arr2[i] = arr[i]; } return arr2; }

var DEBUG = false; // set true to enable console logging.

var E2E_PREFIX = "session.e2e.";
/**
 * Construct a web storage session store, capable of storing account keys,
 * session keys and access tokens.
 * @constructor
 * @param {WebStorage} webStore A web storage implementation, e.g.
 * 'window.localStorage' or 'window.sessionStorage' or a custom implementation.
 * @throws if the supplied 'store' does not meet the Storage interface of the
 * WebStorage API.
 */

function WebStorageSessionStore(webStore) {
  this.store = webStore;

  if (!utils.isFunction(webStore.getItem) || !utils.isFunction(webStore.setItem) || !utils.isFunction(webStore.removeItem) || !utils.isFunction(webStore.key) || typeof webStore.length !== 'number') {
    throw new Error("Supplied webStore does not meet the WebStorage API interface");
  }
}

WebStorageSessionStore.prototype = {
  /**
   * Remove the stored end to end account for the logged-in user.
   */
  removeEndToEndAccount: function removeEndToEndAccount() {
    this.store.removeItem(KEY_END_TO_END_ACCOUNT);
  },

  /**
   * Load the end to end account for the logged-in user.
   * Note that the end-to-end account is now stored in the
   * crypto store rather than here: this remains here so
   * old sessions can be migrated out of the session store.
   * @return {?string} Base64 encoded account.
   */
  getEndToEndAccount: function getEndToEndAccount() {
    return this.store.getItem(KEY_END_TO_END_ACCOUNT);
  },

  /**
   * Retrieves the known devices for all users.
   * @return {object} A map from user ID to map of device ID to keys for the device.
   */
  getAllEndToEndDevices: function getAllEndToEndDevices() {
    var prefix = keyEndToEndDevicesForUser('');
    var devices = {};

    for (var i = 0; i < this.store.length; ++i) {
      var key = this.store.key(i);
      var userId = key.substr(prefix.length);
      if (key.startsWith(prefix)) devices[userId] = getJsonItem(this.store, key);
    }

    return devices;
  },
  getEndToEndDeviceTrackingStatus: function getEndToEndDeviceTrackingStatus() {
    return getJsonItem(this.store, KEY_END_TO_END_DEVICE_LIST_TRACKING_STATUS);
  },

  /**
   * Get the sync token corresponding to the device list.
   *
   * @return {String?} token
   */
  getEndToEndDeviceSyncToken: function getEndToEndDeviceSyncToken() {
    return getJsonItem(this.store, KEY_END_TO_END_DEVICE_SYNC_TOKEN);
  },

  /**
   * Removes all end to end device data from the store
   */
  removeEndToEndDeviceData: function removeEndToEndDeviceData() {
    removeByPrefix(this.store, keyEndToEndDevicesForUser(''));
    removeByPrefix(this.store, KEY_END_TO_END_DEVICE_LIST_TRACKING_STATUS);
    removeByPrefix(this.store, KEY_END_TO_END_DEVICE_SYNC_TOKEN);
  },

  /**
   * Retrieve the end-to-end sessions between the logged-in user and another
   * device.
   * @param {string} deviceKey The public key of the other device.
   * @return {object} A map from sessionId to Base64 end-to-end session.
   */
  getEndToEndSessions: function getEndToEndSessions(deviceKey) {
    return getJsonItem(this.store, keyEndToEndSessions(deviceKey));
  },

  /**
   * Retrieve all end-to-end sessions between the logged-in user and other
   * devices.
   * @return {object} A map of {deviceKey -> {sessionId -> session pickle}}
   */
  getAllEndToEndSessions: function getAllEndToEndSessions() {
    var deviceKeys = getKeysWithPrefix(this.store, keyEndToEndSessions(''));
    var results = {};

    var _iterator = _createForOfIteratorHelper(deviceKeys),
        _step;

    try {
      for (_iterator.s(); !(_step = _iterator.n()).done;) {
        var k = _step.value;
        var unprefixedKey = k.substr(keyEndToEndSessions('').length);
        results[unprefixedKey] = getJsonItem(this.store, k);
      }
    } catch (err) {
      _iterator.e(err);
    } finally {
      _iterator.f();
    }

    return results;
  },

  /**
   * Remove all end-to-end sessions from the store
   * This is used after migrating sessions awat from the sessions store.
   */
  removeAllEndToEndSessions: function removeAllEndToEndSessions() {
    removeByPrefix(this.store, keyEndToEndSessions(''));
  },

  /**
   * Retrieve a list of all known inbound group sessions
   *
   * @return {{senderKey: string, sessionId: string}}
   */
  getAllEndToEndInboundGroupSessionKeys: function getAllEndToEndInboundGroupSessionKeys() {
    var prefix = E2E_PREFIX + 'inboundgroupsessions/';
    var result = [];

    for (var i = 0; i < this.store.length; i++) {
      var key = this.store.key(i);

      if (!key.startsWith(prefix)) {
        continue;
      } // we can't use split, as the components we are trying to split out
      // might themselves contain '/' characters. We rely on the
      // senderKey being a (32-byte) curve25519 key, base64-encoded
      // (hence 43 characters long).


      result.push({
        senderKey: key.substr(prefix.length, 43),
        sessionId: key.substr(prefix.length + 44)
      });
    }

    return result;
  },
  getEndToEndInboundGroupSession: function getEndToEndInboundGroupSession(senderKey, sessionId) {
    var key = keyEndToEndInboundGroupSession(senderKey, sessionId);
    return this.store.getItem(key);
  },
  removeAllEndToEndInboundGroupSessions: function removeAllEndToEndInboundGroupSessions() {
    removeByPrefix(this.store, E2E_PREFIX + 'inboundgroupsessions/');
  },

  /**
   * Get the end-to-end state for all rooms
   * @return {object} roomId -> object with the end-to-end info for the room.
   */
  getAllEndToEndRooms: function getAllEndToEndRooms() {
    var roomKeys = getKeysWithPrefix(this.store, keyEndToEndRoom(''));
    var results = {};

    var _iterator2 = _createForOfIteratorHelper(roomKeys),
        _step2;

    try {
      for (_iterator2.s(); !(_step2 = _iterator2.n()).done;) {
        var k = _step2.value;
        var unprefixedKey = k.substr(keyEndToEndRoom('').length);
        results[unprefixedKey] = getJsonItem(this.store, k);
      }
    } catch (err) {
      _iterator2.e(err);
    } finally {
      _iterator2.f();
    }

    return results;
  },
  removeAllEndToEndRooms: function removeAllEndToEndRooms() {
    removeByPrefix(this.store, keyEndToEndRoom(''));
  },
  setLocalTrustedBackupPubKey: function setLocalTrustedBackupPubKey(pubkey) {
    this.store.setItem(KEY_END_TO_END_TRUSTED_BACKUP_PUBKEY, pubkey);
  },
  // XXX: This store is deprecated really, but added this as a temporary
  // thing until cross-signing lands.
  getLocalTrustedBackupPubKey: function getLocalTrustedBackupPubKey() {
    return this.store.getItem(KEY_END_TO_END_TRUSTED_BACKUP_PUBKEY);
  }
};
var KEY_END_TO_END_ACCOUNT = E2E_PREFIX + "account";
var KEY_END_TO_END_DEVICE_SYNC_TOKEN = E2E_PREFIX + "device_sync_token";
var KEY_END_TO_END_DEVICE_LIST_TRACKING_STATUS = E2E_PREFIX + "device_tracking";
var KEY_END_TO_END_TRUSTED_BACKUP_PUBKEY = E2E_PREFIX + "trusted_backup_pubkey";

function keyEndToEndDevicesForUser(userId) {
  return E2E_PREFIX + "devices/" + userId;
}

function keyEndToEndSessions(deviceKey) {
  return E2E_PREFIX + "sessions/" + deviceKey;
}

function keyEndToEndInboundGroupSession(senderKey, sessionId) {
  return E2E_PREFIX + "inboundgroupsessions/" + senderKey + "/" + sessionId;
}

function keyEndToEndRoom(roomId) {
  return E2E_PREFIX + "rooms/" + roomId;
}

function getJsonItem(store, key) {
  try {
    // if the key is absent, store.getItem() returns null, and
    // JSON.parse(null) === null, so this returns null.
    return JSON.parse(store.getItem(key));
  } catch (e) {
    debuglog("Failed to get key %s: %s", key, e);
    debuglog(e.stack);
  }

  return null;
}

function getKeysWithPrefix(store, prefix) {
  var results = [];

  for (var i = 0; i < store.length; ++i) {
    var key = store.key(i);
    if (key.startsWith(prefix)) results.push(key);
  }

  return results;
}

function removeByPrefix(store, prefix) {
  var toRemove = [];

  for (var i = 0; i < store.length; ++i) {
    var key = store.key(i);
    if (key.startsWith(prefix)) toRemove.push(key);
  }

  for (var _i = 0, _toRemove = toRemove; _i < _toRemove.length; _i++) {
    var _key = _toRemove[_i];
    store.removeItem(_key);
  }
}

function debuglog() {
  if (DEBUG) {
    _logger.logger.log.apply(_logger.logger, arguments);
  }
}