"use strict";

var _interopRequireWildcard = require("@babel/runtime/helpers/interopRequireWildcard");

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.SyncApi = SyncApi;

var _regenerator = _interopRequireDefault(require("@babel/runtime/regenerator"));

var _asyncToGenerator2 = _interopRequireDefault(require("@babel/runtime/helpers/asyncToGenerator"));

var _user = require("./models/user");

var _room = require("./models/room");

var _group = require("./models/group");

var utils = _interopRequireWildcard(require("./utils"));

var _filter = require("./filter");

var _eventTimeline = require("./models/event-timeline");

var _pushprocessor = require("./pushprocessor");

var _logger = require("./logger");

var _errors = require("./errors");

function _createForOfIteratorHelper(o, allowArrayLike) { var it; if (typeof Symbol === "undefined" || o[Symbol.iterator] == null) { if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") { if (it) o = it; var i = 0; var F = function F() {}; return { s: F, n: function n() { if (i >= o.length) return { done: true }; return { done: false, value: o[i++] }; }, e: function e(_e) { throw _e; }, f: F }; } throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); } var normalCompletion = true, didErr = false, err; return { s: function s() { it = o[Symbol.iterator](); }, n: function n() { var step = it.next(); normalCompletion = step.done; return step; }, e: function e(_e2) { didErr = true; err = _e2; }, f: function f() { try { if (!normalCompletion && it["return"] != null) it["return"](); } finally { if (didErr) throw err; } } }; }

function _unsupportedIterableToArray(o, minLen) { if (!o) return; if (typeof o === "string") return _arrayLikeToArray(o, minLen); var n = Object.prototype.toString.call(o).slice(8, -1); if (n === "Object" && o.constructor) n = o.constructor.name; if (n === "Map" || n === "Set") return Array.from(o); if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen); }

function _arrayLikeToArray(arr, len) { if (len == null || len > arr.length) len = arr.length; for (var i = 0, arr2 = new Array(len); i < len; i++) { arr2[i] = arr[i]; } return arr2; }

var DEBUG = true; // /sync requests allow you to set a timeout= but the request may continue
// beyond that and wedge forever, so we need to track how long we are willing
// to keep open the connection. This constant is *ADDED* to the timeout= value
// to determine the max time we're willing to wait.

var BUFFER_PERIOD_MS = 80 * 1000; // Number of consecutive failed syncs that will lead to a syncState of ERROR as opposed
// to RECONNECTING. This is needed to inform the client of server issues when the
// keepAlive is successful but the server /sync fails.

var FAILED_SYNC_ERROR_THRESHOLD = 3;

function getFilterName(userId, suffix) {
  // scope this on the user ID because people may login on many accounts
  // and they all need to be stored!
  return "FILTER_SYNC_" + userId + (suffix ? "_" + suffix : "");
}

function debuglog() {
  if (!DEBUG) {
    return;
  }

  _logger.logger.log.apply(_logger.logger, arguments);
}
/**
 * <b>Internal class - unstable.</b>
 * Construct an entity which is able to sync with a homeserver.
 * @constructor
 * @param {MatrixClient} client The matrix client instance to use.
 * @param {Object} opts Config options
 * @param {module:crypto=} opts.crypto Crypto manager
 * @param {Function=} opts.canResetEntireTimeline A function which is called
 * with a room ID and returns a boolean. It should return 'true' if the SDK can
 * SAFELY remove events from this room. It may not be safe to remove events if
 * there are other references to the timelines for this room.
 * Default: returns false.
 * @param {Boolean=} opts.disablePresence True to perform syncing without automatically
 * updating presence.
 */


function SyncApi(client, opts) {
  this.client = client;
  opts = opts || {};
  opts.initialSyncLimit = opts.initialSyncLimit === undefined ? 8 : opts.initialSyncLimit;
  opts.resolveInvitesToProfiles = opts.resolveInvitesToProfiles || false;
  opts.pollTimeout = opts.pollTimeout || 30 * 1000;
  opts.pendingEventOrdering = opts.pendingEventOrdering || "chronological";

  if (!opts.canResetEntireTimeline) {
    opts.canResetEntireTimeline = function (roomId) {
      return false;
    };
  }

  this.opts = opts;
  this._peekRoomId = null;
  this._currentSyncRequest = null;
  this._syncState = null;
  this._syncStateData = null; // additional data (eg. error object for failed sync)

  this._catchingUp = false;
  this._running = false;
  this._keepAliveTimer = null;
  this._connectionReturnedDefer = null;
  this._notifEvents = []; // accumulator of sync events in the current sync response

  this._failedSyncCount = 0; // Number of consecutive failed /sync requests

  this._storeIsInvalid = false; // flag set if the store needs to be cleared before we can start

  if (client.getNotifTimelineSet()) {
    client.reEmitter.reEmit(client.getNotifTimelineSet(), ["Room.timeline", "Room.timelineReset"]);
  }
}
/**
 * @param {string} roomId
 * @return {Room}
 */


SyncApi.prototype.createRoom = function (roomId) {
  var client = this.client;
  var timelineSupport = client.timelineSupport,
      unstableClientRelationAggregation = client.unstableClientRelationAggregation;
  var room = new _room.Room(roomId, client, client.getUserId(), {
    lazyLoadMembers: this.opts.lazyLoadMembers,
    pendingEventOrdering: this.opts.pendingEventOrdering,
    timelineSupport: timelineSupport,
    unstableClientRelationAggregation: unstableClientRelationAggregation
  });
  client.reEmitter.reEmit(room, ["Room.name", "Room.timeline", "Room.redaction", "Room.redactionCancelled", "Room.receipt", "Room.tags", "Room.timelineReset", "Room.localEchoUpdated", "Room.accountData", "Room.myMembership", "Room.replaceEvent"]);

  this._registerStateListeners(room);

  return room;
};
/**
 * @param {string} groupId
 * @return {Group}
 */


SyncApi.prototype.createGroup = function (groupId) {
  var client = this.client;
  var group = new _group.Group(groupId);
  client.reEmitter.reEmit(group, ["Group.profile", "Group.myMembership"]);
  client.store.storeGroup(group);
  return group;
};
/**
 * @param {Room} room
 * @private
 */


SyncApi.prototype._registerStateListeners = function (room) {
  var client = this.client; // we need to also re-emit room state and room member events, so hook it up
  // to the client now. We need to add a listener for RoomState.members in
  // order to hook them correctly. (TODO: find a better way?)

  client.reEmitter.reEmit(room.currentState, ["RoomState.events", "RoomState.members", "RoomState.newMember"]);
  room.currentState.on("RoomState.newMember", function (event, state, member) {
    member.user = client.getUser(member.userId);
    client.reEmitter.reEmit(member, ["RoomMember.name", "RoomMember.typing", "RoomMember.powerLevel", "RoomMember.membership"]);
  });
};
/**
 * @param {Room} room
 * @private
 */


SyncApi.prototype._deregisterStateListeners = function (room) {
  // could do with a better way of achieving this.
  room.currentState.removeAllListeners("RoomState.events");
  room.currentState.removeAllListeners("RoomState.members");
  room.currentState.removeAllListeners("RoomState.newMember");
};
/**
 * Sync rooms the user has left.
 * @return {Promise} Resolved when they've been added to the store.
 */


SyncApi.prototype.syncLeftRooms = function () {
  var client = this.client;
  var self = this; // grab a filter with limit=1 and include_leave=true

  var filter = new _filter.Filter(this.client.credentials.userId);
  filter.setTimelineLimit(1);
  filter.setIncludeLeaveRooms(true);
  var localTimeoutMs = this.opts.pollTimeout + BUFFER_PERIOD_MS;
  var qps = {
    timeout: 0 // don't want to block since this is a single isolated req

  };
  return client.getOrCreateFilter(getFilterName(client.credentials.userId, "LEFT_ROOMS"), filter).then(function (filterId) {
    qps.filter = filterId;
    return client._http.authedRequest(undefined, "GET", "/sync", qps, undefined, localTimeoutMs);
  }).then(function (data) {
    var leaveRooms = [];

    if (data.rooms && data.rooms.leave) {
      leaveRooms = self._mapSyncResponseToRoomArray(data.rooms.leave);
    }

    var rooms = [];
    leaveRooms.forEach(function (leaveObj) {
      var room = leaveObj.room;
      rooms.push(room);

      if (!leaveObj.isBrandNewRoom) {
        // the intention behind syncLeftRooms is to add in rooms which were
        // *omitted* from the initial /sync. Rooms the user were joined to
        // but then left whilst the app is running will appear in this list
        // and we do not want to bother with them since they will have the
        // current state already (and may get dupe messages if we add
        // yet more timeline events!), so skip them.
        // NB: When we persist rooms to localStorage this will be more
        //     complicated...
        return;
      }

      leaveObj.timeline = leaveObj.timeline || {};

      var timelineEvents = self._mapSyncEventsFormat(leaveObj.timeline, room);

      var stateEvents = self._mapSyncEventsFormat(leaveObj.state, room); // set the back-pagination token. Do this *before* adding any
      // events so that clients can start back-paginating.


      room.getLiveTimeline().setPaginationToken(leaveObj.timeline.prev_batch, _eventTimeline.EventTimeline.BACKWARDS);

      self._processRoomEvents(room, stateEvents, timelineEvents);

      room.recalculate();
      client.store.storeRoom(room);
      client.emit("Room", room);

      self._processEventsForNotifs(room, timelineEvents);
    });
    return rooms;
  });
};
/**
 * Peek into a room. This will result in the room in question being synced so it
 * is accessible via getRooms(). Live updates for the room will be provided.
 * @param {string} roomId The room ID to peek into.
 * @return {Promise} A promise which resolves once the room has been added to the
 * store.
 */


SyncApi.prototype.peek = function (roomId) {
  var self = this;
  var client = this.client;
  this._peekRoomId = roomId;
  return this.client.roomInitialSync(roomId, 20).then(function (response) {
    // make sure things are init'd
    response.messages = response.messages || {};
    response.messages.chunk = response.messages.chunk || [];
    response.state = response.state || [];
    var peekRoom = self.createRoom(roomId); // FIXME: Mostly duplicated from _processRoomEvents but not entirely
    // because "state" in this API is at the BEGINNING of the chunk

    var oldStateEvents = utils.map(utils.deepCopy(response.state), client.getEventMapper());
    var stateEvents = utils.map(response.state, client.getEventMapper());
    var messages = utils.map(response.messages.chunk, client.getEventMapper()); // XXX: copypasted from /sync until we kill off this
    // minging v1 API stuff)
    // handle presence events (User objects)

    if (response.presence && utils.isArray(response.presence)) {
      response.presence.map(client.getEventMapper()).forEach(function (presenceEvent) {
        var user = client.store.getUser(presenceEvent.getContent().user_id);

        if (user) {
          user.setPresenceEvent(presenceEvent);
        } else {
          user = createNewUser(client, presenceEvent.getContent().user_id);
          user.setPresenceEvent(presenceEvent);
          client.store.storeUser(user);
        }

        client.emit("event", presenceEvent);
      });
    } // set the pagination token before adding the events in case people
    // fire off pagination requests in response to the Room.timeline
    // events.


    if (response.messages.start) {
      peekRoom.oldState.paginationToken = response.messages.start;
    } // set the state of the room to as it was after the timeline executes


    peekRoom.oldState.setStateEvents(oldStateEvents);
    peekRoom.currentState.setStateEvents(stateEvents);

    self._resolveInvites(peekRoom);

    peekRoom.recalculate(); // roll backwards to diverge old state. addEventsToTimeline
    // will overwrite the pagination token, so make sure it overwrites
    // it with the right thing.

    peekRoom.addEventsToTimeline(messages.reverse(), true, peekRoom.getLiveTimeline(), response.messages.start);
    client.store.storeRoom(peekRoom);
    client.emit("Room", peekRoom);

    self._peekPoll(peekRoom);

    return peekRoom;
  });
};
/**
 * Stop polling for updates in the peeked room. NOPs if there is no room being
 * peeked.
 */


SyncApi.prototype.stopPeeking = function () {
  this._peekRoomId = null;
};
/**
 * Do a peek room poll.
 * @param {Room} peekRoom
 * @param {string} token from= token
 */


SyncApi.prototype._peekPoll = function (peekRoom, token) {
  if (this._peekRoomId !== peekRoom.roomId) {
    debuglog("Stopped peeking in room %s", peekRoom.roomId);
    return;
  }

  var self = this; // FIXME: gut wrenching; hard-coded timeout values

  this.client._http.authedRequest(undefined, "GET", "/events", {
    room_id: peekRoom.roomId,
    timeout: 30 * 1000,
    from: token
  }, undefined, 50 * 1000).then(function (res) {
    if (self._peekRoomId !== peekRoom.roomId) {
      debuglog("Stopped peeking in room %s", peekRoom.roomId);
      return;
    } // We have a problem that we get presence both from /events and /sync
    // however, /sync only returns presence for users in rooms
    // you're actually joined to.
    // in order to be sure to get presence for all of the users in the
    // peeked room, we handle presence explicitly here. This may result
    // in duplicate presence events firing for some users, which is a
    // performance drain, but such is life.
    // XXX: copypasted from /sync until we can kill this minging v1 stuff.


    res.chunk.filter(function (e) {
      return e.type === "m.presence";
    }).map(self.client.getEventMapper()).forEach(function (presenceEvent) {
      var user = self.client.store.getUser(presenceEvent.getContent().user_id);

      if (user) {
        user.setPresenceEvent(presenceEvent);
      } else {
        user = createNewUser(self.client, presenceEvent.getContent().user_id);
        user.setPresenceEvent(presenceEvent);
        self.client.store.storeUser(user);
      }

      self.client.emit("event", presenceEvent);
    }); // strip out events which aren't for the given room_id (e.g presence)
    // and also ephemeral events (which we're assuming is anything without
    // and event ID because the /events API doesn't separate them).

    var events = res.chunk.filter(function (e) {
      return e.room_id === peekRoom.roomId && e.event_id;
    }).map(self.client.getEventMapper());
    peekRoom.addLiveEvents(events);

    self._peekPoll(peekRoom, res.end);
  }, function (err) {
    _logger.logger.error("[%s] Peek poll failed: %s", peekRoom.roomId, err);

    setTimeout(function () {
      self._peekPoll(peekRoom, token);
    }, 30 * 1000);
  });
};
/**
 * Returns the current state of this sync object
 * @see module:client~MatrixClient#event:"sync"
 * @return {?String}
 */


SyncApi.prototype.getSyncState = function () {
  return this._syncState;
};
/**
 * Returns the additional data object associated with
 * the current sync state, or null if there is no
 * such data.
 * Sync errors, if available, are put in the 'error' key of
 * this object.
 * @return {?Object}
 */


SyncApi.prototype.getSyncStateData = function () {
  return this._syncStateData;
};

SyncApi.prototype.recoverFromSyncStartupError = /*#__PURE__*/function () {
  var _ref = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee(savedSyncPromise, err) {
    var keepaliveProm;
    return _regenerator["default"].wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            _context.next = 2;
            return savedSyncPromise;

          case 2:
            keepaliveProm = this._startKeepAlives();

            this._updateSyncState("ERROR", {
              error: err
            });

            _context.next = 6;
            return keepaliveProm;

          case 6:
          case "end":
            return _context.stop();
        }
      }
    }, _callee, this);
  }));

  return function (_x, _x2) {
    return _ref.apply(this, arguments);
  };
}();
/**
 * Is the lazy loading option different than in previous session?
 * @param {bool} lazyLoadMembers current options for lazy loading
 * @return {bool} whether or not the option has changed compared to the previous session */


SyncApi.prototype._wasLazyLoadingToggled = /*#__PURE__*/function () {
  var _ref2 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee2(lazyLoadMembers) {
    var lazyLoadMembersBefore, isStoreNewlyCreated, prevClientOptions;
    return _regenerator["default"].wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            lazyLoadMembers = !!lazyLoadMembers; // assume it was turned off before
            // if we don't know any better

            lazyLoadMembersBefore = false;
            _context2.next = 4;
            return this.client.store.isNewlyCreated();

          case 4:
            isStoreNewlyCreated = _context2.sent;

            if (isStoreNewlyCreated) {
              _context2.next = 11;
              break;
            }

            _context2.next = 8;
            return this.client.store.getClientOptions();

          case 8:
            prevClientOptions = _context2.sent;

            if (prevClientOptions) {
              lazyLoadMembersBefore = !!prevClientOptions.lazyLoadMembers;
            }

            return _context2.abrupt("return", lazyLoadMembersBefore !== lazyLoadMembers);

          case 11:
            return _context2.abrupt("return", false);

          case 12:
          case "end":
            return _context2.stop();
        }
      }
    }, _callee2, this);
  }));

  return function (_x3) {
    return _ref2.apply(this, arguments);
  };
}();

SyncApi.prototype._shouldAbortSync = function (error) {
  if (error.errcode === "M_UNKNOWN_TOKEN") {
    // The logout already happened, we just need to stop.
    _logger.logger.warn("Token no longer valid - assuming logout");

    this.stop();
    return true;
  }

  return false;
};
/**
 * Main entry point
 */


SyncApi.prototype.sync = function () {
  var _this = this;

  var client = this.client;
  var self = this;
  this._running = true;

  if (global.window) {
    this._onOnlineBound = this._onOnline.bind(this);
    global.window.addEventListener("online", this._onOnlineBound, false);
  }

  var savedSyncPromise = Promise.resolve();
  var savedSyncToken = null; // We need to do one-off checks before we can begin the /sync loop.
  // These are:
  //   1) We need to get push rules so we can check if events should bing as we get
  //      them from /sync.
  //   2) We need to get/create a filter which we can use for /sync.
  //   3) We need to check the lazy loading option matches what was used in the
  //       stored sync. If it doesn't, we can't use the stored sync.

  function getPushRules() {
    return _getPushRules.apply(this, arguments);
  }

  function _getPushRules() {
    _getPushRules = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee4() {
      var result;
      return _regenerator["default"].wrap(function _callee4$(_context4) {
        while (1) {
          switch (_context4.prev = _context4.next) {
            case 0:
              _context4.prev = 0;
              debuglog("Getting push rules...");
              _context4.next = 4;
              return client.getPushRules();

            case 4:
              result = _context4.sent;
              debuglog("Got push rules");
              client.pushRules = result;
              _context4.next = 19;
              break;

            case 9:
              _context4.prev = 9;
              _context4.t0 = _context4["catch"](0);

              _logger.logger.error("Getting push rules failed", _context4.t0);

              if (!self._shouldAbortSync(_context4.t0)) {
                _context4.next = 14;
                break;
              }

              return _context4.abrupt("return");

            case 14:
              // wait for saved sync to complete before doing anything else,
              // otherwise the sync state will end up being incorrect
              debuglog("Waiting for saved sync before retrying push rules...");
              _context4.next = 17;
              return self.recoverFromSyncStartupError(savedSyncPromise, _context4.t0);

            case 17:
              getPushRules();
              return _context4.abrupt("return");

            case 19:
              checkLazyLoadStatus(); // advance to the next stage

            case 20:
            case "end":
              return _context4.stop();
          }
        }
      }, _callee4, null, [[0, 9]]);
    }));
    return _getPushRules.apply(this, arguments);
  }

  function buildDefaultFilter() {
    var filter = new _filter.Filter(client.credentials.userId);
    filter.setTimelineLimit(self.opts.initialSyncLimit);
    return filter;
  }

  var checkLazyLoadStatus = /*#__PURE__*/function () {
    var _ref3 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee3() {
      var supported, shouldClear, reason, error;
      return _regenerator["default"].wrap(function _callee3$(_context3) {
        while (1) {
          switch (_context3.prev = _context3.next) {
            case 0:
              debuglog("Checking lazy load status...");

              if (_this.opts.lazyLoadMembers && client.isGuest()) {
                _this.opts.lazyLoadMembers = false;
              }

              if (!_this.opts.lazyLoadMembers) {
                _context3.next = 8;
                break;
              }

              debuglog("Checking server lazy load support...");
              _context3.next = 6;
              return client.doesServerSupportLazyLoading();

            case 6:
              supported = _context3.sent;

              if (supported) {
                debuglog("Enabling lazy load on sync filter...");

                if (!_this.opts.filter) {
                  _this.opts.filter = buildDefaultFilter();
                }

                _this.opts.filter.setLazyLoadMembers(true);
              } else {
                debuglog("LL: lazy loading requested but not supported " + "by server, so disabling");
                _this.opts.lazyLoadMembers = false;
              }

            case 8:
              // need to vape the store when enabling LL and wasn't enabled before
              debuglog("Checking whether lazy loading has changed in store...");
              _context3.next = 11;
              return _this._wasLazyLoadingToggled(_this.opts.lazyLoadMembers);

            case 11:
              shouldClear = _context3.sent;

              if (!shouldClear) {
                _context3.next = 19;
                break;
              }

              _this._storeIsInvalid = true;
              reason = _errors.InvalidStoreError.TOGGLED_LAZY_LOADING;
              error = new _errors.InvalidStoreError(reason, !!_this.opts.lazyLoadMembers);

              _this._updateSyncState("ERROR", {
                error: error
              }); // bail out of the sync loop now: the app needs to respond to this error.
              // we leave the state as 'ERROR' which isn't great since this normally means
              // we're retrying. The client must be stopped before clearing the stores anyway
              // so the app should stop the client, clear the store and start it again.


              _logger.logger.warn("InvalidStoreError: store is not usable: stopping sync.");

              return _context3.abrupt("return");

            case 19:
              if (_this.opts.lazyLoadMembers && _this.opts.crypto) {
                _this.opts.crypto.enableLazyLoading();
              }

              _context3.prev = 20;
              debuglog("Storing client options...");
              _context3.next = 24;
              return _this.client._storeClientOptions();

            case 24:
              debuglog("Stored client options");
              _context3.next = 31;
              break;

            case 27:
              _context3.prev = 27;
              _context3.t0 = _context3["catch"](20);

              _logger.logger.error("Storing client options failed", _context3.t0);

              throw _context3.t0;

            case 31:
              getFilter(); // Now get the filter and start syncing

            case 32:
            case "end":
              return _context3.stop();
          }
        }
      }, _callee3, null, [[20, 27]]);
    }));

    return function checkLazyLoadStatus() {
      return _ref3.apply(this, arguments);
    };
  }();

  function getFilter() {
    return _getFilter.apply(this, arguments);
  }

  function _getFilter() {
    _getFilter = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee5() {
      var filter, filterId;
      return _regenerator["default"].wrap(function _callee5$(_context5) {
        while (1) {
          switch (_context5.prev = _context5.next) {
            case 0:
              debuglog("Getting filter...");

              if (self.opts.filter) {
                filter = self.opts.filter;
              } else {
                filter = buildDefaultFilter();
              }

              _context5.prev = 2;
              _context5.next = 5;
              return client.getOrCreateFilter(getFilterName(client.credentials.userId), filter);

            case 5:
              filterId = _context5.sent;
              _context5.next = 18;
              break;

            case 8:
              _context5.prev = 8;
              _context5.t0 = _context5["catch"](2);

              _logger.logger.error("Getting filter failed", _context5.t0);

              if (!self._shouldAbortSync(_context5.t0)) {
                _context5.next = 13;
                break;
              }

              return _context5.abrupt("return");

            case 13:
              // wait for saved sync to complete before doing anything else,
              // otherwise the sync state will end up being incorrect
              debuglog("Waiting for saved sync before retrying filter...");
              _context5.next = 16;
              return self.recoverFromSyncStartupError(savedSyncPromise, _context5.t0);

            case 16:
              getFilter();
              return _context5.abrupt("return");

            case 18:
              // reset the notifications timeline to prepare it to paginate from
              // the current point in time.
              // The right solution would be to tie /sync pagination tokens into
              // /notifications API somehow.
              client.resetNotifTimelineSet();

              if (self._currentSyncRequest === null) {
                // Send this first sync request here so we can then wait for the saved
                // sync data to finish processing before we process the results of this one.
                debuglog("Sending first sync request...");
                self._currentSyncRequest = self._doSyncRequest({
                  filterId: filterId
                }, savedSyncToken);
              } // Now wait for the saved sync to finish...


              debuglog("Waiting for saved sync before starting sync processing...");
              _context5.next = 23;
              return savedSyncPromise;

            case 23:
              self._sync({
                filterId: filterId
              });

            case 24:
            case "end":
              return _context5.stop();
          }
        }
      }, _callee5, null, [[2, 8]]);
    }));
    return _getFilter.apply(this, arguments);
  }

  if (client.isGuest()) {
    // no push rules for guests, no access to POST filter for guests.
    self._sync({});
  } else {
    // Pull the saved sync token out first, before the worker starts sending
    // all the sync data which could take a while. This will let us send our
    // first incremental sync request before we've processed our saved data.
    debuglog("Getting saved sync token...");
    savedSyncPromise = client.store.getSavedSyncToken().then(function (tok) {
      debuglog("Got saved sync token");
      savedSyncToken = tok;
      debuglog("Getting saved sync...");
      return client.store.getSavedSync();
    }).then(function (savedSync) {
      debuglog("Got reply from saved sync, exists? ".concat(!!savedSync));

      if (savedSync) {
        return self._syncFromCache(savedSync);
      }
    })["catch"](function (err) {
      _logger.logger.error("Getting saved sync failed", err);
    }); // Now start the first incremental sync request: this can also
    // take a while so if we set it going now, we can wait for it
    // to finish while we process our saved sync data.

    getPushRules();
  }
};
/**
 * Stops the sync object from syncing.
 */


SyncApi.prototype.stop = function () {
  debuglog("SyncApi.stop");

  if (global.window) {
    global.window.removeEventListener("online", this._onOnlineBound, false);
    this._onOnlineBound = undefined;
  }

  this._running = false;

  if (this._currentSyncRequest) {
    this._currentSyncRequest.abort();
  }

  if (this._keepAliveTimer) {
    clearTimeout(this._keepAliveTimer);
    this._keepAliveTimer = null;
  }
};
/**
 * Retry a backed off syncing request immediately. This should only be used when
 * the user <b>explicitly</b> attempts to retry their lost connection.
 * @return {boolean} True if this resulted in a request being retried.
 */


SyncApi.prototype.retryImmediately = function () {
  if (!this._connectionReturnedDefer) {
    return false;
  }

  this._startKeepAlives(0);

  return true;
};
/**
 * Process a single set of cached sync data.
 * @param {Object} savedSync a saved sync that was persisted by a store. This
 * should have been acquired via client.store.getSavedSync().
 */


SyncApi.prototype._syncFromCache = /*#__PURE__*/function () {
  var _ref4 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee6(savedSync) {
    var nextSyncToken, syncEventData, data;
    return _regenerator["default"].wrap(function _callee6$(_context6) {
      while (1) {
        switch (_context6.prev = _context6.next) {
          case 0:
            debuglog("sync(): not doing HTTP hit, instead returning stored /sync data");
            nextSyncToken = savedSync.nextBatch; // Set sync token for future incremental syncing

            this.client.store.setSyncToken(nextSyncToken); // No previous sync, set old token to null

            syncEventData = {
              oldSyncToken: null,
              nextSyncToken: nextSyncToken,
              catchingUp: false,
              fromCache: true
            };
            data = {
              next_batch: nextSyncToken,
              rooms: savedSync.roomsData,
              groups: savedSync.groupsData,
              account_data: {
                events: savedSync.accountData
              }
            };
            _context6.prev = 5;
            _context6.next = 8;
            return this._processSyncResponse(syncEventData, data);

          case 8:
            _context6.next = 13;
            break;

          case 10:
            _context6.prev = 10;
            _context6.t0 = _context6["catch"](5);

            _logger.logger.error("Error processing cached sync", _context6.t0.stack || _context6.t0);

          case 13:
            // Don't emit a prepared if we've bailed because the store is invalid:
            // in this case the client will not be usable until stopped & restarted
            // so this would be useless and misleading.
            if (!this._storeIsInvalid) {
              this._updateSyncState("PREPARED", syncEventData);
            }

          case 14:
          case "end":
            return _context6.stop();
        }
      }
    }, _callee6, this, [[5, 10]]);
  }));

  return function (_x4) {
    return _ref4.apply(this, arguments);
  };
}();
/**
 * Invoke me to do /sync calls
 * @param {Object} syncOptions
 * @param {string} syncOptions.filterId
 * @param {boolean} syncOptions.hasSyncedBefore
 */


SyncApi.prototype._sync = /*#__PURE__*/function () {
  var _ref5 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee7(syncOptions) {
    var client, syncToken, data, syncEventData;
    return _regenerator["default"].wrap(function _callee7$(_context7) {
      while (1) {
        switch (_context7.prev = _context7.next) {
          case 0:
            client = this.client;

            if (this._running) {
              _context7.next = 6;
              break;
            }

            debuglog("Sync no longer running: exiting.");

            if (this._connectionReturnedDefer) {
              this._connectionReturnedDefer.reject();

              this._connectionReturnedDefer = null;
            }

            this._updateSyncState("STOPPED");

            return _context7.abrupt("return");

          case 6:
            syncToken = client.store.getSyncToken();
            _context7.prev = 7;

            //debuglog('Starting sync since=' + syncToken);
            if (this._currentSyncRequest === null) {
              this._currentSyncRequest = this._doSyncRequest(syncOptions, syncToken);
            }

            _context7.next = 11;
            return this._currentSyncRequest;

          case 11:
            data = _context7.sent;
            _context7.next = 18;
            break;

          case 14:
            _context7.prev = 14;
            _context7.t0 = _context7["catch"](7);

            this._onSyncError(_context7.t0, syncOptions);

            return _context7.abrupt("return");

          case 18:
            _context7.prev = 18;
            this._currentSyncRequest = null;
            return _context7.finish(18);

          case 21:
            //debuglog('Completed sync, next_batch=' + data.next_batch);
            // set the sync token NOW *before* processing the events. We do this so
            // if something barfs on an event we can skip it rather than constantly
            // polling with the same token.
            client.store.setSyncToken(data.next_batch); // Reset after a successful sync

            this._failedSyncCount = 0;
            _context7.next = 25;
            return client.store.setSyncData(data);

          case 25:
            syncEventData = {
              oldSyncToken: syncToken,
              nextSyncToken: data.next_batch,
              catchingUp: this._catchingUp
            };

            if (!this.opts.crypto) {
              _context7.next = 29;
              break;
            }

            _context7.next = 29;
            return this.opts.crypto.onSyncWillProcess(syncEventData);

          case 29:
            _context7.prev = 29;
            _context7.next = 32;
            return this._processSyncResponse(syncEventData, data);

          case 32:
            _context7.next = 38;
            break;

          case 34:
            _context7.prev = 34;
            _context7.t1 = _context7["catch"](29);

            // log the exception with stack if we have it, else fall back
            // to the plain description
            _logger.logger.error("Caught /sync error", _context7.t1.stack || _context7.t1); // Emit the exception for client handling


            this.client.emit("sync.unexpectedError", _context7.t1);

          case 38:
            // update this as it may have changed
            syncEventData.catchingUp = this._catchingUp; // emit synced events

            if (!syncOptions.hasSyncedBefore) {
              this._updateSyncState("PREPARED", syncEventData);

              syncOptions.hasSyncedBefore = true;
            } // tell the crypto module to do its processing. It may block (to do a
            // /keys/changes request).


            if (!this.opts.crypto) {
              _context7.next = 43;
              break;
            }

            _context7.next = 43;
            return this.opts.crypto.onSyncCompleted(syncEventData);

          case 43:
            // keep emitting SYNCING -> SYNCING for clients who want to do bulk updates
            this._updateSyncState("SYNCING", syncEventData);

            if (!client.store.wantsSave()) {
              _context7.next = 49;
              break;
            }

            if (!this.opts.crypto) {
              _context7.next = 48;
              break;
            }

            _context7.next = 48;
            return this.opts.crypto.saveDeviceList(0);

          case 48:
            // tell databases that everything is now in a consistent state and can be saved.
            client.store.save();

          case 49:
            // Begin next sync
            this._sync(syncOptions);

          case 50:
          case "end":
            return _context7.stop();
        }
      }
    }, _callee7, this, [[7, 14, 18, 21], [29, 34]]);
  }));

  return function (_x5) {
    return _ref5.apply(this, arguments);
  };
}();

SyncApi.prototype._doSyncRequest = function (syncOptions, syncToken) {
  var qps = this._getSyncParams(syncOptions, syncToken);

  return this.client._http.authedRequest(undefined, "GET", "/sync", qps, undefined, qps.timeout + BUFFER_PERIOD_MS);
};

SyncApi.prototype._getSyncParams = function (syncOptions, syncToken) {
  var pollTimeout = this.opts.pollTimeout;

  if (this.getSyncState() !== 'SYNCING' || this._catchingUp) {
    // unless we are happily syncing already, we want the server to return
    // as quickly as possible, even if there are no events queued. This
    // serves two purposes:
    //
    // * When the connection dies, we want to know asap when it comes back,
    //   so that we can hide the error from the user. (We don't want to
    //   have to wait for an event or a timeout).
    //
    // * We want to know if the server has any to_device messages queued up
    //   for us. We do that by calling it with a zero timeout until it
    //   doesn't give us any more to_device messages.
    this._catchingUp = true;
    pollTimeout = 0;
  }

  var filterId = syncOptions.filterId;

  if (this.client.isGuest() && !filterId) {
    filterId = this._getGuestFilter();
  }

  var qps = {
    filter: filterId,
    timeout: pollTimeout
  };

  if (this.opts.disablePresence) {
    qps.set_presence = "offline";
  }

  if (syncToken) {
    qps.since = syncToken;
  } else {
    // use a cachebuster for initialsyncs, to make sure that
    // we don't get a stale sync
    // (https://github.com/vector-im/vector-web/issues/1354)
    qps._cacheBuster = Date.now();
  }

  if (this.getSyncState() == 'ERROR' || this.getSyncState() == 'RECONNECTING') {
    // we think the connection is dead. If it comes back up, we won't know
    // about it till /sync returns. If the timeout= is high, this could
    // be a long time. Set it to 0 when doing retries so we don't have to wait
    // for an event or a timeout before emiting the SYNCING event.
    qps.timeout = 0;
  }

  return qps;
};

SyncApi.prototype._onSyncError = function (err, syncOptions) {
  var _this2 = this;

  if (!this._running) {
    debuglog("Sync no longer running: exiting");

    if (this._connectionReturnedDefer) {
      this._connectionReturnedDefer.reject();

      this._connectionReturnedDefer = null;
    }

    this._updateSyncState("STOPPED");

    return;
  }

  _logger.logger.error("/sync error %s", err);

  _logger.logger.error(err);

  if (this._shouldAbortSync(err)) {
    return;
  }

  this._failedSyncCount++;

  _logger.logger.log('Number of consecutive failed sync requests:', this._failedSyncCount);

  debuglog("Starting keep-alive"); // Note that we do *not* mark the sync connection as
  // lost yet: we only do this if a keepalive poke
  // fails, since long lived HTTP connections will
  // go away sometimes and we shouldn't treat this as
  // erroneous. We set the state to 'reconnecting'
  // instead, so that clients can observe this state
  // if they wish.

  this._startKeepAlives().then(function (connDidFail) {
    // Only emit CATCHUP if we detected a connectivity error: if we didn't,
    // it's quite likely the sync will fail again for the same reason and we
    // want to stay in ERROR rather than keep flip-flopping between ERROR
    // and CATCHUP.
    if (connDidFail && _this2.getSyncState() === 'ERROR') {
      _this2._updateSyncState("CATCHUP", {
        oldSyncToken: null,
        nextSyncToken: null,
        catchingUp: true
      });
    }

    _this2._sync(syncOptions);
  });

  this._currentSyncRequest = null; // Transition from RECONNECTING to ERROR after a given number of failed syncs

  this._updateSyncState(this._failedSyncCount >= FAILED_SYNC_ERROR_THRESHOLD ? "ERROR" : "RECONNECTING", {
    error: err
  });
};
/**
 * Process data returned from a sync response and propagate it
 * into the model objects
 *
 * @param {Object} syncEventData Object containing sync tokens associated with this sync
 * @param {Object} data The response from /sync
 */


SyncApi.prototype._processSyncResponse = /*#__PURE__*/function () {
  var _ref6 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee10(syncEventData, data) {
    var client, self, events, prevEventsMap, cancelledKeyVerificationTxns, inviteRooms, joinRooms, leaveRooms, currentCount, unusedFallbackKeys;
    return _regenerator["default"].wrap(function _callee10$(_context10) {
      while (1) {
        switch (_context10.prev = _context10.next) {
          case 0:
            client = this.client;
            self = this; // data looks like:
            // {
            //    next_batch: $token,
            //    presence: { events: [] },
            //    account_data: { events: [] },
            //    device_lists: { changed: ["@user:server", ... ]},
            //    to_device: { events: [] },
            //    device_one_time_keys_count: { signed_curve25519: 42 },
            //    rooms: {
            //      invite: {
            //        $roomid: {
            //          invite_state: { events: [] }
            //        }
            //      },
            //      join: {
            //        $roomid: {
            //          state: { events: [] },
            //          timeline: { events: [], prev_batch: $token, limited: true },
            //          ephemeral: { events: [] },
            //          summary: {
            //             m.heroes: [ $user_id ],
            //             m.joined_member_count: $count,
            //             m.invited_member_count: $count
            //          },
            //          account_data: { events: [] },
            //          unread_notifications: {
            //              highlight_count: 0,
            //              notification_count: 0,
            //          }
            //        }
            //      },
            //      leave: {
            //        $roomid: {
            //          state: { events: [] },
            //          timeline: { events: [], prev_batch: $token }
            //        }
            //      }
            //    },
            //    groups: {
            //        invite: {
            //            $groupId: {
            //                inviter: $inviter,
            //                profile: {
            //                    avatar_url: $avatarUrl,
            //                    name: $groupName,
            //                },
            //            },
            //        },
            //        join: {},
            //        leave: {},
            //    },
            // }
            // TODO-arch:
            // - Each event we pass through needs to be emitted via 'event', can we
            //   do this in one place?
            // - The isBrandNewRoom boilerplate is boilerplatey.
            // handle presence events (User objects)

            if (data.presence && utils.isArray(data.presence.events)) {
              data.presence.events.map(client.getEventMapper()).forEach(function (presenceEvent) {
                var user = client.store.getUser(presenceEvent.getSender());

                if (user) {
                  user.setPresenceEvent(presenceEvent);
                } else {
                  user = createNewUser(client, presenceEvent.getSender());
                  user.setPresenceEvent(presenceEvent);
                  client.store.storeUser(user);
                }

                client.emit("event", presenceEvent);
              });
            } // handle non-room account_data


            if (data.account_data && utils.isArray(data.account_data.events)) {
              events = data.account_data.events.map(client.getEventMapper());
              prevEventsMap = events.reduce(function (m, c) {
                m[c.getId()] = client.store.getAccountData(c.getType());
                return m;
              }, {});
              client.store.storeAccountDataEvents(events);
              events.forEach(function (accountDataEvent) {
                // Honour push rules that come down the sync stream but also
                // honour push rules that were previously cached. Base rules
                // will be updated when we receive push rules via getPushRules
                // (see SyncApi.prototype.sync) before syncing over the network.
                if (accountDataEvent.getType() === 'm.push_rules') {
                  var rules = accountDataEvent.getContent();
                  client.pushRules = _pushprocessor.PushProcessor.rewriteDefaultRules(rules);
                }

                var prevEvent = prevEventsMap[accountDataEvent.getId()];
                client.emit("accountData", accountDataEvent, prevEvent);
                return accountDataEvent;
              });
            } // handle to-device events


            if (data.to_device && utils.isArray(data.to_device.events) && data.to_device.events.length > 0) {
              cancelledKeyVerificationTxns = [];
              data.to_device.events.map(client.getEventMapper()).map(function (toDeviceEvent) {
                // map is a cheap inline forEach
                // We want to flag m.key.verification.start events as cancelled
                // if there's an accompanying m.key.verification.cancel event, so
                // we pull out the transaction IDs from the cancellation events
                // so we can flag the verification events as cancelled in the loop
                // below.
                if (toDeviceEvent.getType() === "m.key.verification.cancel") {
                  var txnId = toDeviceEvent.getContent()['transaction_id'];

                  if (txnId) {
                    cancelledKeyVerificationTxns.push(txnId);
                  }
                } // as mentioned above, .map is a cheap inline forEach, so return
                // the unmodified event.


                return toDeviceEvent;
              }).forEach(function (toDeviceEvent) {
                var content = toDeviceEvent.getContent();

                if (toDeviceEvent.getType() == "m.room.message" && content.msgtype == "m.bad.encrypted") {
                  // the mapper already logged a warning.
                  _logger.logger.log('Ignoring undecryptable to-device event from ' + toDeviceEvent.getSender());

                  return;
                }

                if (toDeviceEvent.getType() === "m.key.verification.start" || toDeviceEvent.getType() === "m.key.verification.request") {
                  var txnId = content['transaction_id'];

                  if (cancelledKeyVerificationTxns.includes(txnId)) {
                    toDeviceEvent.flagCancelled();
                  }
                }

                client.emit("toDeviceEvent", toDeviceEvent);
              });
            } else {
              // no more to-device events: we can stop polling with a short timeout.
              this._catchingUp = false;
            }

            if (data.groups) {
              if (data.groups.invite) {
                this._processGroupSyncEntry(data.groups.invite, 'invite');
              }

              if (data.groups.join) {
                this._processGroupSyncEntry(data.groups.join, 'join');
              }

              if (data.groups.leave) {
                this._processGroupSyncEntry(data.groups.leave, 'leave');
              }
            } // the returned json structure is a bit crap, so make it into a
            // nicer form (array) after applying sanity to make sure we don't fail
            // on missing keys (on the off chance)


            inviteRooms = [];
            joinRooms = [];
            leaveRooms = [];

            if (data.rooms) {
              if (data.rooms.invite) {
                inviteRooms = this._mapSyncResponseToRoomArray(data.rooms.invite);
              }

              if (data.rooms.join) {
                joinRooms = this._mapSyncResponseToRoomArray(data.rooms.join);
              }

              if (data.rooms.leave) {
                leaveRooms = this._mapSyncResponseToRoomArray(data.rooms.leave);
              }
            }

            this._notifEvents = []; // Handle invites

            inviteRooms.forEach(function (inviteObj) {
              var room = inviteObj.room;

              var stateEvents = self._mapSyncEventsFormat(inviteObj.invite_state, room);

              self._processRoomEvents(room, stateEvents);

              if (inviteObj.isBrandNewRoom) {
                room.recalculate();
                client.store.storeRoom(room);
                client.emit("Room", room);
              }

              stateEvents.forEach(function (e) {
                client.emit("event", e);
              });
              room.updateMyMembership("invite");
            }); // Handle joins

            _context10.next = 14;
            return utils.promiseMapSeries(joinRooms, /*#__PURE__*/function () {
              var _ref7 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee9(joinObj) {
                var room, stateEvents, timelineEvents, ephemeralEvents, accountDataEvents, encrypted, limited, i, eventId, processRoomEvent, _processRoomEvent;

                return _regenerator["default"].wrap(function _callee9$(_context9) {
                  while (1) {
                    switch (_context9.prev = _context9.next) {
                      case 0:
                        _processRoomEvent = function _processRoomEvent3() {
                          _processRoomEvent = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee8(e) {
                            var user;
                            return _regenerator["default"].wrap(function _callee8$(_context8) {
                              while (1) {
                                switch (_context8.prev = _context8.next) {
                                  case 0:
                                    client.emit("event", e);

                                    if (!(e.isState() && e.getType() == "m.room.encryption" && self.opts.crypto)) {
                                      _context8.next = 4;
                                      break;
                                    }

                                    _context8.next = 4;
                                    return self.opts.crypto.onCryptoEvent(e);

                                  case 4:
                                    if (e.isState() && e.getType() === "im.vector.user_status") {
                                      user = client.store.getUser(e.getStateKey());

                                      if (user) {
                                        user._unstable_updateStatusMessage(e);
                                      } else {
                                        user = createNewUser(client, e.getStateKey());

                                        user._unstable_updateStatusMessage(e);

                                        client.store.storeUser(user);
                                      }
                                    }

                                  case 5:
                                  case "end":
                                    return _context8.stop();
                                }
                              }
                            }, _callee8);
                          }));
                          return _processRoomEvent.apply(this, arguments);
                        };

                        processRoomEvent = function _processRoomEvent2(_x9) {
                          return _processRoomEvent.apply(this, arguments);
                        };

                        room = joinObj.room;
                        stateEvents = self._mapSyncEventsFormat(joinObj.state, room);
                        timelineEvents = self._mapSyncEventsFormat(joinObj.timeline, room);
                        ephemeralEvents = self._mapSyncEventsFormat(joinObj.ephemeral);
                        accountDataEvents = self._mapSyncEventsFormat(joinObj.account_data); // we do this first so it's correct when any of the events fire

                        if (joinObj.unread_notifications) {
                          room.setUnreadNotificationCount('total', joinObj.unread_notifications.notification_count); // We track unread notifications ourselves in encrypted rooms, so don't
                          // bother setting it here. We trust our calculations better than the
                          // server's for this case, and therefore will assume that our non-zero
                          // count is accurate.

                          encrypted = client.isRoomEncrypted(room.roomId);

                          if (!encrypted || encrypted && room.getUnreadNotificationCount('highlight') <= 0) {
                            room.setUnreadNotificationCount('highlight', joinObj.unread_notifications.highlight_count);
                          }
                        }

                        joinObj.timeline = joinObj.timeline || {};

                        if (!joinObj.isBrandNewRoom) {
                          _context9.next = 13;
                          break;
                        }

                        // set the back-pagination token. Do this *before* adding any
                        // events so that clients can start back-paginating.
                        room.getLiveTimeline().setPaginationToken(joinObj.timeline.prev_batch, _eventTimeline.EventTimeline.BACKWARDS);
                        _context9.next = 27;
                        break;

                      case 13:
                        if (!joinObj.timeline.limited) {
                          _context9.next = 27;
                          break;
                        }

                        limited = true; // we've got a limited sync, so we *probably* have a gap in the
                        // timeline, so should reset. But we might have been peeking or
                        // paginating and already have some of the events, in which
                        // case we just want to append any subsequent events to the end
                        // of the existing timeline.
                        //
                        // This is particularly important in the case that we already have
                        // *all* of the events in the timeline - in that case, if we reset
                        // the timeline, we'll end up with an entirely empty timeline,
                        // which we'll try to paginate but not get any new events (which
                        // will stop us linking the empty timeline into the chain).
                        //

                        i = timelineEvents.length - 1;

                      case 16:
                        if (!(i >= 0)) {
                          _context9.next = 26;
                          break;
                        }

                        eventId = timelineEvents[i].getId();

                        if (!room.getTimelineForEvent(eventId)) {
                          _context9.next = 23;
                          break;
                        }

                        debuglog("Already have event " + eventId + " in limited " + "sync - not resetting");
                        limited = false; // we might still be missing some of the events before i;
                        // we don't want to be adding them to the end of the
                        // timeline because that would put them out of order.

                        timelineEvents.splice(0, i); // XXX: there's a problem here if the skipped part of the
                        // timeline modifies the state set in stateEvents, because
                        // we'll end up using the state from stateEvents rather
                        // than the later state from timelineEvents. We probably
                        // need to wind stateEvents forward over the events we're
                        // skipping.

                        return _context9.abrupt("break", 26);

                      case 23:
                        i--;
                        _context9.next = 16;
                        break;

                      case 26:
                        if (limited) {
                          self._deregisterStateListeners(room);

                          room.resetLiveTimeline(joinObj.timeline.prev_batch, self.opts.canResetEntireTimeline(room.roomId) ? null : syncEventData.oldSyncToken); // We have to assume any gap in any timeline is
                          // reason to stop incrementally tracking notifications and
                          // reset the timeline.

                          client.resetNotifTimelineSet();

                          self._registerStateListeners(room);
                        }

                      case 27:
                        self._processRoomEvents(room, stateEvents, timelineEvents, syncEventData.fromCache); // set summary after processing events,
                        // because it will trigger a name calculation
                        // which needs the room state to be up to date


                        if (joinObj.summary) {
                          room.setSummary(joinObj.summary);
                        } // we deliberately don't add ephemeral events to the timeline


                        room.addEphemeralEvents(ephemeralEvents); // we deliberately don't add accountData to the timeline

                        room.addAccountData(accountDataEvents);
                        room.recalculate();

                        if (joinObj.isBrandNewRoom) {
                          client.store.storeRoom(room);
                          client.emit("Room", room);
                        }

                        self._processEventsForNotifs(room, timelineEvents);

                        _context9.next = 36;
                        return utils.promiseMapSeries(stateEvents, processRoomEvent);

                      case 36:
                        _context9.next = 38;
                        return utils.promiseMapSeries(timelineEvents, processRoomEvent);

                      case 38:
                        ephemeralEvents.forEach(function (e) {
                          client.emit("event", e);
                        });
                        accountDataEvents.forEach(function (e) {
                          client.emit("event", e);
                        });
                        room.updateMyMembership("join");

                      case 41:
                      case "end":
                        return _context9.stop();
                    }
                  }
                }, _callee9);
              }));

              return function (_x8) {
                return _ref7.apply(this, arguments);
              };
            }());

          case 14:
            // Handle leaves (e.g. kicked rooms)
            leaveRooms.forEach(function (leaveObj) {
              var room = leaveObj.room;

              var stateEvents = self._mapSyncEventsFormat(leaveObj.state, room);

              var timelineEvents = self._mapSyncEventsFormat(leaveObj.timeline, room);

              var accountDataEvents = self._mapSyncEventsFormat(leaveObj.account_data);

              self._processRoomEvents(room, stateEvents, timelineEvents);

              room.addAccountData(accountDataEvents);
              room.recalculate();

              if (leaveObj.isBrandNewRoom) {
                client.store.storeRoom(room);
                client.emit("Room", room);
              }

              self._processEventsForNotifs(room, timelineEvents);

              stateEvents.forEach(function (e) {
                client.emit("event", e);
              });
              timelineEvents.forEach(function (e) {
                client.emit("event", e);
              });
              accountDataEvents.forEach(function (e) {
                client.emit("event", e);
              });
              room.updateMyMembership("leave");
            }); // update the notification timeline, if appropriate.
            // we only do this for live events, as otherwise we can't order them sanely
            // in the timeline relative to ones paginated in by /notifications.
            // XXX: we could fix this by making EventTimeline support chronological
            // ordering... but it doesn't, right now.

            if (syncEventData.oldSyncToken && this._notifEvents.length) {
              this._notifEvents.sort(function (a, b) {
                return a.getTs() - b.getTs();
              });

              this._notifEvents.forEach(function (event) {
                client.getNotifTimelineSet().addLiveEvent(event);
              });
            } // Handle device list updates


            if (!data.device_lists) {
              _context10.next = 22;
              break;
            }

            if (!this.opts.crypto) {
              _context10.next = 22;
              break;
            }

            _context10.next = 20;
            return this.opts.crypto.handleDeviceListChanges(syncEventData, data.device_lists);

          case 20:
            _context10.next = 22;
            break;

          case 22:
            // Handle one_time_keys_count
            if (this.opts.crypto && data.device_one_time_keys_count) {
              currentCount = data.device_one_time_keys_count.signed_curve25519 || 0;
              this.opts.crypto.updateOneTimeKeyCount(currentCount);
            }

            if (this.opts.crypto && data["org.matrix.msc2732.device_unused_fallback_key_types"]) {
              // The presence of device_unused_fallback_key_types indicates that the
              // server supports fallback keys. If there's no unused
              // signed_curve25519 fallback key we need a new one.
              unusedFallbackKeys = data["org.matrix.msc2732.device_unused_fallback_key_types"];
              this.opts.crypto.setNeedsNewFallback(unusedFallbackKeys instanceof Array && !unusedFallbackKeys.includes("signed_curve25519"));
            }

          case 24:
          case "end":
            return _context10.stop();
        }
      }
    }, _callee10, this);
  }));

  return function (_x6, _x7) {
    return _ref6.apply(this, arguments);
  };
}();
/**
 * Starts polling the connectivity check endpoint
 * @param {number} delay How long to delay until the first poll.
 *        defaults to a short, randomised interval (to prevent
 *        tightlooping if /versions succeeds but /sync etc. fail).
 * @return {promise} which resolves once the connection returns
 */


SyncApi.prototype._startKeepAlives = function (delay) {
  if (delay === undefined) {
    delay = 2000 + Math.floor(Math.random() * 5000);
  }

  if (this._keepAliveTimer !== null) {
    clearTimeout(this._keepAliveTimer);
  }

  var self = this;

  if (delay > 0) {
    self._keepAliveTimer = setTimeout(self._pokeKeepAlive.bind(self), delay);
  } else {
    self._pokeKeepAlive();
  }

  if (!this._connectionReturnedDefer) {
    this._connectionReturnedDefer = utils.defer();
  }

  return this._connectionReturnedDefer.promise;
};
/**
 * Make a dummy call to /_matrix/client/versions, to see if the HS is
 * reachable.
 *
 * On failure, schedules a call back to itself. On success, resolves
 * this._connectionReturnedDefer.
 *
 * @param {bool} connDidFail True if a connectivity failure has been detected. Optional.
 */


SyncApi.prototype._pokeKeepAlive = function (connDidFail) {
  if (connDidFail === undefined) connDidFail = false;
  var self = this;

  function success() {
    clearTimeout(self._keepAliveTimer);

    if (self._connectionReturnedDefer) {
      self._connectionReturnedDefer.resolve(connDidFail);

      self._connectionReturnedDefer = null;
    }
  }

  this.client._http.request(undefined, // callback
  "GET", "/_matrix/client/versions", undefined, // queryParams
  undefined, // data
  {
    prefix: '',
    localTimeoutMs: 15 * 1000
  }).then(function () {
    success();
  }, function (err) {
    if (err.httpStatus == 400 || err.httpStatus == 404) {
      // treat this as a success because the server probably just doesn't
      // support /versions: point is, we're getting a response.
      // We wait a short time though, just in case somehow the server
      // is in a mode where it 400s /versions responses and sync etc.
      // responses fail, this will mean we don't hammer in a loop.
      self._keepAliveTimer = setTimeout(success, 2000);
    } else {
      connDidFail = true;
      self._keepAliveTimer = setTimeout(self._pokeKeepAlive.bind(self, connDidFail), 5000 + Math.floor(Math.random() * 5000)); // A keepalive has failed, so we emit the
      // error state (whether or not this is the
      // first failure).
      // Note we do this after setting the timer:
      // this lets the unit tests advance the mock
      // clock when they get the error.

      self._updateSyncState("ERROR", {
        error: err
      });
    }
  });
};
/**
 * @param {Object} groupsSection Groups section object, eg. response.groups.invite
 * @param {string} sectionName Which section this is ('invite', 'join' or 'leave')
 */


SyncApi.prototype._processGroupSyncEntry = function (groupsSection, sectionName) {
  // Processes entries from 'groups' section of the sync stream
  for (var _i = 0, _Object$keys = Object.keys(groupsSection); _i < _Object$keys.length; _i++) {
    var groupId = _Object$keys[_i];
    var groupInfo = groupsSection[groupId];
    var group = this.client.store.getGroup(groupId);
    var isBrandNew = group === null;

    if (group === null) {
      group = this.createGroup(groupId);
    }

    if (groupInfo.profile) {
      group.setProfile(groupInfo.profile.name, groupInfo.profile.avatar_url);
    }

    if (groupInfo.inviter) {
      group.setInviter({
        userId: groupInfo.inviter
      });
    }

    group.setMyMembership(sectionName);

    if (isBrandNew) {
      // Now we've filled in all the fields, emit the Group event
      this.client.emit("Group", group);
    }
  }
};
/**
 * @param {Object} obj
 * @return {Object[]}
 */


SyncApi.prototype._mapSyncResponseToRoomArray = function (obj) {
  // Maps { roomid: {stuff}, roomid: {stuff} }
  // to
  // [{stuff+Room+isBrandNewRoom}, {stuff+Room+isBrandNewRoom}]
  var client = this.client;
  var self = this;
  return utils.keys(obj).map(function (roomId) {
    var arrObj = obj[roomId];
    var room = client.store.getRoom(roomId);
    var isBrandNewRoom = false;

    if (!room) {
      room = self.createRoom(roomId);
      isBrandNewRoom = true;
    }

    arrObj.room = room;
    arrObj.isBrandNewRoom = isBrandNewRoom;
    return arrObj;
  });
};
/**
 * @param {Object} obj
 * @param {Room} room
 * @return {MatrixEvent[]}
 */


SyncApi.prototype._mapSyncEventsFormat = function (obj, room) {
  if (!obj || !utils.isArray(obj.events)) {
    return [];
  }

  var mapper = this.client.getEventMapper();
  return obj.events.map(function (e) {
    if (room) {
      e.room_id = room.roomId;
    }

    return mapper(e);
  });
};
/**
 * @param {Room} room
 */


SyncApi.prototype._resolveInvites = function (room) {
  if (!room || !this.opts.resolveInvitesToProfiles) {
    return;
  }

  var client = this.client; // For each invited room member we want to give them a displayname/avatar url
  // if they have one (the m.room.member invites don't contain this).

  room.getMembersWithMembership("invite").forEach(function (member) {
    if (member._requestedProfileInfo) {
      return;
    }

    member._requestedProfileInfo = true; // try to get a cached copy first.

    var user = client.getUser(member.userId);
    var promise;

    if (user) {
      promise = Promise.resolve({
        avatar_url: user.avatarUrl,
        displayname: user.displayName
      });
    } else {
      promise = client.getProfileInfo(member.userId);
    }

    promise.then(function (info) {
      // slightly naughty by doctoring the invite event but this means all
      // the code paths remain the same between invite/join display name stuff
      // which is a worthy trade-off for some minor pollution.
      var inviteEvent = member.events.member;

      if (inviteEvent.getContent().membership !== "invite") {
        // between resolving and now they have since joined, so don't clobber
        return;
      }

      inviteEvent.getContent().avatar_url = info.avatar_url;
      inviteEvent.getContent().displayname = info.displayname; // fire listeners

      member.setMembershipEvent(inviteEvent, room.currentState);
    }, function (err) {// OH WELL.
    });
  });
};
/**
 * @param {Room} room
 * @param {MatrixEvent[]} stateEventList A list of state events. This is the state
 * at the *START* of the timeline list if it is supplied.
 * @param {MatrixEvent[]} [timelineEventList] A list of timeline events. Lower index
 * @param {boolean} fromCache whether the sync response came from cache
 * is earlier in time. Higher index is later.
 */


SyncApi.prototype._processRoomEvents = function (room, stateEventList, timelineEventList, fromCache) {
  // If there are no events in the timeline yet, initialise it with
  // the given state events
  var liveTimeline = room.getLiveTimeline();
  var timelineWasEmpty = liveTimeline.getEvents().length == 0;

  if (timelineWasEmpty) {
    // Passing these events into initialiseState will freeze them, so we need
    // to compute and cache the push actions for them now, otherwise sync dies
    // with an attempt to assign to read only property.
    // XXX: This is pretty horrible and is assuming all sorts of behaviour from
    // these functions that it shouldn't be. We should probably either store the
    // push actions cache elsewhere so we can freeze MatrixEvents, or otherwise
    // find some solution where MatrixEvents are immutable but allow for a cache
    // field.
    var _iterator = _createForOfIteratorHelper(stateEventList),
        _step;

    try {
      for (_iterator.s(); !(_step = _iterator.n()).done;) {
        var ev = _step.value;
        this.client.getPushActionsForEvent(ev);
      }
    } catch (err) {
      _iterator.e(err);
    } finally {
      _iterator.f();
    }

    liveTimeline.initialiseState(stateEventList);
  }

  this._resolveInvites(room); // recalculate the room name at this point as adding events to the timeline
  // may make notifications appear which should have the right name.
  // XXX: This looks suspect: we'll end up recalculating the room once here
  // and then again after adding events (_processSyncResponse calls it after
  // calling us) even if no state events were added. It also means that if
  // one of the room events in timelineEventList is something that needs
  // a recalculation (like m.room.name) we won't recalculate until we've
  // finished adding all the events, which will cause the notification to have
  // the old room name rather than the new one.


  room.recalculate(); // If the timeline wasn't empty, we process the state events here: they're
  // defined as updates to the state before the start of the timeline, so this
  // starts to roll the state forward.
  // XXX: That's what we *should* do, but this can happen if we were previously
  // peeking in a room, in which case we obviously do *not* want to add the
  // state events here onto the end of the timeline. Historically, the js-sdk
  // has just set these new state events on the old and new state. This seems
  // very wrong because there could be events in the timeline that diverge the
  // state, in which case this is going to leave things out of sync. However,
  // for now I think it;s best to behave the same as the code has done previously.

  if (!timelineWasEmpty) {
    // XXX: As above, don't do this...
    //room.addLiveEvents(stateEventList || []);
    // Do this instead...
    room.oldState.setStateEvents(stateEventList || []);
    room.currentState.setStateEvents(stateEventList || []);
  } // execute the timeline events. This will continue to diverge the current state
  // if the timeline has any state events in it.
  // This also needs to be done before running push rules on the events as they need
  // to be decorated with sender etc.


  room.addLiveEvents(timelineEventList || [], null, fromCache);
};
/**
 * Takes a list of timelineEvents and adds and adds to _notifEvents
 * as appropriate.
 * This must be called after the room the events belong to has been stored.
 *
 * @param {Room} room
 * @param {MatrixEvent[]} [timelineEventList] A list of timeline events. Lower index
 * is earlier in time. Higher index is later.
 */


SyncApi.prototype._processEventsForNotifs = function (room, timelineEventList) {
  // gather our notifications into this._notifEvents
  if (this.client.getNotifTimelineSet()) {
    for (var i = 0; i < timelineEventList.length; i++) {
      var pushActions = this.client.getPushActionsForEvent(timelineEventList[i]);

      if (pushActions && pushActions.notify && pushActions.tweaks && pushActions.tweaks.highlight) {
        this._notifEvents.push(timelineEventList[i]);
      }
    }
  }
};
/**
 * @return {string}
 */


SyncApi.prototype._getGuestFilter = function () {
  var guestRooms = this.client._guestRooms; // FIXME: horrible gut-wrenching

  if (!guestRooms) {
    return "{}";
  } // we just need to specify the filter inline if we're a guest because guests
  // can't create filters.


  return JSON.stringify({
    room: {
      timeline: {
        limit: 20
      }
    }
  });
};
/**
 * Sets the sync state and emits an event to say so
 * @param {String} newState The new state string
 * @param {Object} data Object of additional data to emit in the event
 */


SyncApi.prototype._updateSyncState = function (newState, data) {
  var old = this._syncState;
  this._syncState = newState;
  this._syncStateData = data;
  this.client.emit("sync", this._syncState, old, data);
};
/**
 * Event handler for the 'online' event
 * This event is generally unreliable and precise behaviour
 * varies between browsers, so we poll for connectivity too,
 * but this might help us reconnect a little faster.
 */


SyncApi.prototype._onOnline = function () {
  debuglog("Browser thinks we are back online");

  this._startKeepAlives(0);
};

function createNewUser(client, userId) {
  var user = new _user.User(userId);
  client.reEmitter.reEmit(user, ["User.avatarUrl", "User.displayName", "User.presence", "User.currentlyActive", "User.lastPresenceTs"]);
  return user;
}